from datetime import datetime, timezone
from os import stat

import pytest
from flow.record.fieldtypes import path

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.apps.ssh.putty import PuTTYPlugin, construct_public_key
from tests._utils import absolute_path


def test_putty_plugin_ssh_host_keys_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    ssh_host_keys = absolute_path("_data/plugins/apps/ssh/putty/sshhostkeys")
    fs_unix.map_file("/root/.putty/sshhostkeys", ssh_host_keys)

    target_unix_users.add_plugin(PuTTYPlugin)
    records = list(target_unix_users.putty.known_hosts())

    assert len(records) == 3

    assert records[0].mtime_ts is not None
    assert records[0].host == "192.168.123.130"
    assert records[0].port == 22

    assert records[0].key_type == "ssh-ed25519"
    assert records[0].public_key == "AAAAC3NzaC1lZDI1NTE5AAAAIHUl23i/4p/7xcZnNPDK+Dr+A539zpEEXutrm/tESFYq"
    assert records[0].fingerprint.md5 == "16da68cfc0e1b7954c84147a385be2b6"
    assert records[0].fingerprint.sha1 == "11e647366ef9b74feb55da7f8507a7180123eed0"
    assert records[0].fingerprint.sha256 == "0ecaca9db7d166fd55f84ad775074bd6c743ae2d26e6dac10e60e88efbfcc01d"

    assert records[0].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[0].username == "root"

    assert records[1].mtime_ts is not None
    assert records[1].host == "example.com"
    assert records[1].port == 22

    assert records[1].key_type == "ecdsa-sha2-nistp256"
    assert (
        records[1].public_key
        == "AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBNhJOmUcnvcVIizc67M3bS0/2Gz43YPHYCRaogqAgHvgVUartyfsl5nZBoFR3FiUc9kTJJUPybWbqXaGLEthjfM="  # noqa E501
    )
    assert records[1].fingerprint.md5 == "e23736ef80f12b672cf217c390a1acfb"
    assert records[1].fingerprint.sha1 == "93a03a809a02091aab194b04c836c37298c0602a"
    assert records[1].fingerprint.sha256 == "53d7097c310ce329d554dfb533c81f81d955facae8f46e9d7cfa4db58304c674"

    assert records[1].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[1].username == "root"

    assert records[2].mtime_ts is not None
    assert records[2].host == "1.2.3.4"
    assert records[2].port == 1234

    assert records[2].key_type == "rsa2"
    assert (
        records[2].public_key
        == "AAAAB3NzaC1yc2EAAAADAQABAAABgQCmYvCKueiA3lyhZuW0OAObOZvf0g957H0wGhGi3BruZSDpa9UQhsVNtgPZfuyx/yYWUIhQ4VSz+qw00IhZJt9myRNrXCxIABM/qubTghrdjRU+ydb0J9uTBIRz+ys/0dr0dg2Gc7C5w9+E3EpRit6VCZfTi8mNYZi+/GC12VYIDsqY9/4D4xloHJP+fs1mCNnY12VtJuIWw281fKaTCxm3H95NOBbAE6sAwr8H5lMLU34D4DcA5ZIG5F48yDmYUkILtN4qLAZMl4ZlQ6KpcxMYCN6DPj2NwR8JHzr+mkwjmGqbs8tjYD/KRYL2sYhK6Fdx7wvw5djQeVTEnhjxoeiCYNFAGkzB9BfeP4N0K2rp6F3NktFU4WIc6nLsX5G1LEWuAvAxwg+v6y4YzmpXHP+WrRHKyhS+B64aLpmD7AAJzPwSyqtt5+8SY3z7EMCpYBGz8uhdYzaY0U+qWFHtIfI5GLrl2akYBzH9t/YPAF6wdTuSh4jxVe2IVrd1x7g23Zs="  # noqa E501
    )
    assert records[2].fingerprint.md5 == "539539a9dec6afa0c8d9cdbb9357ed5e"
    assert records[2].fingerprint.sha1 == "c0f0c0a51ed7ae4c2b271f0426554f837f9508d0"
    assert records[2].fingerprint.sha256 == "c2d4bd813cdf1eec38ea8c7bbf0c56b129f9db3e78627d5b7ebfae8db60c5eb6"

    assert records[2].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[2].username == "root"


def test_putty_plugin_ssh_host_keys_windows(
    target_win_users: Target, fs_win: VirtualFilesystem, hive_hku: VirtualHive
) -> None:
    key_name = "Software\\SimonTatham\\PuTTY\\SshHostKeys"
    key = VirtualKey(hive_hku, key_name)
    key.add_value(
        "ssh-ed25519@22:192.168.123.130",
        VirtualValue(
            hive_hku,
            "ssh-ed25519@22:192.168.123.130",
            "0x12812d95024c0d8683879fd38f977e9caa7e733334f75965bc50eba6e872e70a,0x2a564844fb9b6beb5e0491cefd9d03fe3af8caf03467c6c5fb9fe2bf78db2575",  # noqa E501
        ),
    )
    hive_hku.map_key(key_name, key)
    target_win_users.add_plugin(PuTTYPlugin)

    records = list(target_win_users.putty.known_hosts())

    assert len(records) == 1

    assert records[0].host == "192.168.123.130"
    assert records[0].port == 22
    assert records[0].key_type == "ssh-ed25519"
    assert records[0].public_key == "AAAAC3NzaC1lZDI1NTE5AAAAIHUl23i/4p/7xcZnNPDK+Dr+A539zpEEXutrm/tESFYq"
    assert records[0].path == path.from_windows("Software\\SimonTatham\\PuTTY\\SshHostKeys")


def test_putty_plugin_saved_sessions_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    sessions_folder = absolute_path("_data/plugins/apps/ssh/putty/sessions")
    fs_unix.map_dir("/root/.putty/sessions", sessions_folder)

    target_unix_users.add_plugin(PuTTYPlugin)
    records = list(target_unix_users.putty.sessions())

    assert len(records) == 1

    assert records[0].ts == datetime.fromtimestamp(
        stat(absolute_path("_data/plugins/apps/ssh/putty/sessions/example-saved-session")).st_mtime,
        tz=timezone.utc,
    )
    assert records[0].session_name == "example-saved-session"
    assert records[0].protocol == "ssh"
    assert records[0].host == "192.168.123.130"
    assert records[0].user == "saveduser"
    assert records[0].port == 22
    assert records[0].remote_command == ""
    assert records[0].port_forward == ""
    assert records[0].manual_ssh_host_keys == ""
    assert records[0].path == path.from_posix("/root/.putty/sessions/example-saved-session")
    assert records[0].username == "root"


def test_putty_plugin_saved_sessions_windows(
    target_win_users: Target, fs_win: VirtualFilesystem, hive_hku: VirtualHive
) -> None:
    key_name = "Software\\SimonTatham\\PuTTY\\Sessions\\example-saved-session"
    key = VirtualKey(hive_hku, key_name)
    key.add_value("HostName", VirtualValue(hive_hku, "HostName", "example.com"))
    key.add_value("UserName", VirtualValue(hive_hku, "UserName", "user"))
    key.add_value("Protocol", VirtualValue(hive_hku, "Protocol", "ssh"))
    key.add_value("PortNumber", VirtualValue(hive_hku, "PortNumber", 22))
    key.add_value("RemoteCommand", VirtualValue(hive_hku, "RemoteCommand", ""))
    key.add_value("PortForwardings", VirtualValue(hive_hku, "PortForwardings", ""))
    key.add_value("SSHManualHostKeys", VirtualValue(hive_hku, "SSHManualHostKeys", ""))
    hive_hku.map_key(key_name, key)
    target_win_users.add_plugin(PuTTYPlugin)

    records = list(target_win_users.putty.sessions())

    assert len(records) == 1

    assert records[0].session_name == "example-saved-session"
    assert records[0].protocol == "ssh"
    assert records[0].host == "example.com"
    assert records[0].user == "user"
    assert records[0].port == 22
    assert records[0].remote_command == ""
    assert records[0].port_forward == ""
    assert records[0].manual_ssh_host_keys == ""
    assert records[0].path == path.from_windows("Software\\SimonTatham\\PuTTY\\Sessions\\example-saved-session")


def test_construct_public_key() -> None:
    with pytest.raises(ValueError):
        construct_public_key(None, None)

    assert construct_public_key("unsupported", "some-iv") == (
        "some-iv",
        (None, None, None),
    )



import csv
import io
import ipaddress
import struct
from pathlib import Path
from typing import Iterator

from dissect.util.ts import wintimestamp

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

SEPLogRecord = TargetRecordDescriptor(
    "application/av/symantec/sep/log",
    [
        ("datetime", "ts"),
        ("string", "virus"),
        ("string", "user"),
        ("path", "source_file"),
        ("string", "action_taken"),
        ("string", "virus_type"),
        ("varint", "scan_id"),
        ("varint", "quarantine_id"),
        ("varint", "virus_id"),
        ("varint", "depth"),
        ("boolean", "still_infected"),
        ("boolean", "quarantined"),
        ("boolean", "compressed"),
        ("boolean", "cleanable"),
        ("boolean", "deletable"),
        ("varint", "confidence"),
        ("varint", "prevalence"),
        ("varint", "risk"),
        ("uri", "download_url"),
        ("varint", "line_no"),
    ],
)

SEPFirewallRecord = TargetRecordDescriptor(
    "application/av/symantec/sep/firewall",
    [
        ("datetime", "ts"),
        ("string", "protocol"),
        ("net.ipaddress", "local_ip"),
        ("net.ipaddress", "remote_ip"),
        ("net.ipaddress", "local_ip6"),
        ("net.ipaddress", "remote_ip6"),
        ("varint", "local_port"),
        ("varint", "remote_port"),
        ("boolean", "outbound"),
        ("datetime", "begin_time"),
        ("datetime", "end_time"),
        ("varint", "repetition"),
        ("boolean", "blocked"),
        ("string", "severity"),
        ("varint", "rule_id"),
        ("string", "remote_host"),
        ("string", "rule_name"),
        ("path", "application"),
        ("string", "user"),
        ("varint", "line_no"),
    ],
)


class SymantecPlugin(Plugin):
    """Symantec Endpoint Security Suite Plugin, based on https://malwaremaloney.blogspot.com/2021/01/"""

    __namespace__ = "symantec"

    LOG_SEP_AV = "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Logs/AV/*"
    LOG_SEP_NET = "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Logs/tralog.log"

    LOGS = [LOG_SEP_AV, LOG_SEP_NET]

    # Special values
    MARKER_INFECTION = 5
    QUARANTINE_SUCCESS = 2
    CLEANABLE = 0
    DELETABLE = 4
    STILL_INFECTED = 1
    OUTBOUND = 2
    BLOCKED = 1
    COMPRESSED = 1

    TCP_INIT = 301
    TCP_CLOSE = 304
    UDP_DATA = 302

    # Log fields
    AV_TIMESTAMP = 0
    AV_EVENT = 1
    AV_USER = 5
    AV_VIRUS = 6
    AV_FILE = 7
    AV_ACTION_TAKEN = 10
    AV_VIRUS_TYPE = 11
    AV_SCAN_ID = 14
    AV_EVENT_DATA = 17
    AV_QUARANTINE_ID = 18
    AV_VIRUS_ID = 19
    AV_QUARANTINE_STATUS = 20
    AV_COMPRESSED = 23
    AV_DEPTH = 24
    AV_STILL_INFECTED = 25
    AV_CLEANABLE = 28
    AV_DELETABLE = 29
    AV_CONFIDENCE = 65
    AV_PREVALENCE = 67
    AV_DOWNLOADED_FROM = 68
    AV_RISK = 71

    # Firwall fields
    FW_TIMESTAMP = 1
    FW_PROTOCOL = 2
    FW_LOCAL_IP = 3
    FW_REMOTE_IP = 4
    FW_LOCAL_PORT = 5
    FW_REMOTE_PORT = 6
    FW_DIRECTION = 7
    FW_BEGIN_TIME = 8
    FW_END_TIME = 9
    FW_REPETITION = 10
    FW_ACTION = 11
    FW_SEVERITY = 13
    FW_RULE_ID = 14
    FW_REMOTE_HOST_NAME = 15
    FW_RULE_NAME = 16
    FW_APPLICATION = 17
    FW_LOCATION = 20
    FW_USER = 21
    FW_LOCAL_IP6 = 25
    FW_REMOTE_IP6 = 26

    PROTOCOL = {
        301: "TCP initiated",
        302: "UDP datagram",
        303: "Ping request",
        304: "TCP completed",
        305: "Traffic (other)",
        306: "ICMPv4 packet",
        307: "Ethernet packet",
        308: "IP packet",
        309: "ICMPv6 packet",
    }

    SEVERITY = ["Critical"] * 4 + ["Major"] * 4 + ["Minor"] * 4 + ["Info"] * 4

    ACTION = {
        1: "Quarantine",
        2: "Rename",
        3: "Delete",
        4: "Leave Alone",
        5: "Clean",
        6: "Remove Macros",
        7: "Save file as...",
        8: "Sent to backend",
        9: "Restore from Quarantine",
        11: "Undo Action",
        12: "Error",
        13: "Backup to quarantine (backup view)",
        14: "Pending Analysis",
        15: "Partially Fixed",
        16: "Terminate Process Required",
        17: "Exclude from Scanning",
        18: "Reboot Processing",
        19: "Clean by Deletion",
        20: "Access Denied",
        21: "TERMINATE PROCESS ONLY",
        22: "NO REPAIR",
        23: "FAIL",
        24: "RUN POWERTOOL",
        25: "NO REPAIR POWERTOOL",
        110: "INTERESTING PROCESS CAL",
        111: "INTERESTING PROCESS DETECTED",
        1000: "INTERESTING PROCESS HASHED DETECTED",
        1001: "DNS HOST FILE EXCEPTION",
    }

    VIRUS_TYPE = {
        48: "Heuristic",
        64: "Reputation",
        80: "Hack Tools",
        96: "Spyware",
        112: "Trackware",
        128: "Dialers",
        144: "Remote Access",
        160: "Adware",
        176: "Joke Programs",
        224: "Heuristic Application",
        256: "Test",
    }

    def __init__(self, target: Target) -> None:
        super().__init__(target)
        self.codepage = self.target.codepage or "ascii"

    def check_compatible(self) -> None:
        for log_file in self.LOGS:
            if list(self.target.fs.glob(log_file)):
                return
        raise UnsupportedPluginError("No Symantec SEP logs found")

    def _fw_cell(self, line: list, cell_id: int) -> str:
        return line[cell_id].decode("utf-8")

    def _fw_hex_cell(self, line: list, cell_id: int) -> int:
        return int(self._fw_cell(line, cell_id), 16)

    def _septime(self, ts: str) -> str:
        # Date/Time stored as 12 digit hex number (6 pairs of 2 digits)
        # Offsets for year (+1970) and month (+1), no TZ
        year = int(ts[0:2], 16)
        month = int(ts[2:4], 16)
        day = int(ts[4:6], 16)
        hour = int(ts[6:8], 16)
        minute = int(ts[8:10], 16)
        second = int(ts[10:12], 16)
        return f"{year+1970}-{month+1:0>2}-{day:0>2}T{hour:0>2}:{minute:0>2}:{second:0>2}"

    def _fw_line(self, line: bytes, line_no: int) -> Iterator[SEPFirewallRecord]:
        try:
            cells = line.split(b"\t")
            protocol_id = self._fw_hex_cell(cells, self.FW_PROTOCOL)
            tcp_udp = protocol_id in [self.TCP_INIT, self.TCP_CLOSE, self.UDP_DATA]

            yield SEPFirewallRecord(
                ts=wintimestamp(self._fw_hex_cell(cells, self.FW_TIMESTAMP)),
                protocol=self.PROTOCOL.get(protocol_id, "Unknown"),
                local_ip=ipaddress.ip_address(struct.pack("<i", self._fw_hex_cell(cells, self.FW_LOCAL_IP))),
                remote_ip=ipaddress.ip_address(struct.pack("<i", self._fw_hex_cell(cells, self.FW_REMOTE_IP))),
                local_ip6=ipaddress.ip_address(ipaddress.v6_int_to_packed(self._fw_hex_cell(cells, self.FW_LOCAL_IP6))),
                remote_ip6=ipaddress.ip_address(
                    ipaddress.v6_int_to_packed(self._fw_hex_cell(cells, self.FW_REMOTE_IP6))
                ),
                local_port=self._fw_hex_cell(cells, self.FW_LOCAL_PORT) if tcp_udp else None,
                remote_port=self._fw_hex_cell(cells, self.FW_REMOTE_PORT) if tcp_udp else None,
                outbound=self._fw_hex_cell(cells, self.FW_DIRECTION) == self.OUTBOUND,
                begin_time=wintimestamp(self._fw_hex_cell(cells, self.FW_BEGIN_TIME)),
                end_time=wintimestamp(self._fw_hex_cell(cells, self.FW_END_TIME)),
                repetition=self._fw_hex_cell(cells, self.FW_REPETITION),
                blocked=self._fw_hex_cell(cells, self.FW_ACTION) == self.BLOCKED,
                severity=self.SEVERITY[self._fw_hex_cell(cells, self.FW_SEVERITY)],
                rule_id=self._fw_hex_cell(cells, self.FW_RULE_ID),
                remote_host=self._fw_cell(cells, self.FW_REMOTE_HOST_NAME),
                rule_name=self._fw_cell(cells, self.FW_RULE_NAME),
                application=Path(self._fw_cell(cells, self.FW_APPLICATION)),
                user=self._fw_cell(cells, self.FW_USER),
                line_no=line_no,
                _target=self.target,
            )
        except Exception as error:
            self.target.log.warning(f"Error: {error} on firewall log line: {line_no}.")

    def _line(self, line: str, line_no: int) -> Iterator[SEPLogRecord]:
        try:
            rows = csv.reader(io.StringIO(line))

            for cells in rows:
                if int(cells[self.AV_EVENT]) != self.MARKER_INFECTION:
                    return

                yield SEPLogRecord(
                    ts=self._septime(cells[self.AV_TIMESTAMP]),
                    virus=cells[self.AV_VIRUS],
                    virus_type=self.VIRUS_TYPE.get(int(cells[self.AV_VIRUS_TYPE]), "Unknown"),
                    user=cells[self.AV_USER],
                    source_file=cells[self.AV_FILE],
                    action_taken=self.ACTION.get(int(cells[self.AV_ACTION_TAKEN]), "Unknown"),
                    scan_id=cells[self.AV_SCAN_ID],
                    quarantine_id=cells[self.AV_QUARANTINE_ID],
                    virus_id=cells[self.AV_VIRUS_ID],
                    quarantined=int(cells[self.AV_QUARANTINE_STATUS]) == self.QUARANTINE_SUCCESS,
                    compressed=int(cells[self.AV_COMPRESSED]) == self.COMPRESSED,
                    depth=cells[self.AV_DEPTH],
                    still_infected=(int(cells[self.AV_STILL_INFECTED]) == self.STILL_INFECTED),
                    cleanable=int(cells[self.AV_CLEANABLE]) == self.DELETABLE,
                    deletable=int(cells[self.AV_DELETABLE]) == self.CLEANABLE,
                    confidence=cells[self.AV_CONFIDENCE],
                    prevalence=cells[self.AV_PREVALENCE],
                    risk=cells[self.AV_RISK],
                    download_url=cells[self.AV_DOWNLOADED_FROM],
                    line_no=line_no,
                    _target=self.target,
                )
        except Exception as error:
            self.target.log.warning(f"Error: {error} on log line: {line_no}.")

    @export(record=SEPLogRecord)
    def logs(self) -> Iterator[SEPLogRecord]:
        """Return log records.

        Yields SEPLogRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp associated with the event.
            virus (string): Name of the virus.
            user (string): Name of the user associated with the event.
            source_file (path): File that contains the virus.
            action_taken (string): Action taken by SEP.
            virus_type (string): Description of the type of virus.
            scan_id (varint): ID of the scan associated with the event.
            event_data (string): String or bytes from a virus event.
            quarantine_id (varint): ID associated with the quarantined virus.
            still_infected (boolean): Whether the system is still infected.
            quarantined (boolean): True if the virus has been quarantined succesfully.
            compressed (boolean): True if the virus was in a compressed file.
            depth (varint): How many layers of compression the virus was hidden in.
            cleanable (boolean): Whether the virus is cleanable.
            deletable (boolean): Whether the virus can be deleted.
            confidence (varint): Confidence level about threat verdict (higher is more confident).
            prevalence (varint): Prevalence of the threat (higher is more prevalent).
            risk (varint): Risk level of the threat (1-4, higher is more dangerous, 0 = unknown).
            download_url (uri): Source of the virus (if available).
            line_no (varint): Reference line number in log file.
        """
        for log_path in self.target.fs.glob(self.LOG_SEP_AV):
            with self.target.fs.path(log_path).open("rt", 0, self.codepage) as csv_file:
                line_no = 0
                while line := csv_file.readline():
                    line_no += 1
                    yield from self._line(line, line_no)

    @export(record=SEPFirewallRecord)
    def firewall(self) -> Iterator[SEPFirewallRecord]:
        """Return log firewall records.

        Yields SEPFirewallRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp associated with the event.
            protocol (string): Protocol name associated with the firewall record.
            local_ip ("net.ipaddress"): Local IP address associated with the event.
            remote_ip ("net.ipaddress"): Remote IP address associated with the event.
            local_ip6 ("net.ipaddress"): Local IPv6 address associated with the event.
            remote_ip6 ("net.ipaddress"): Remote IPv6 address associated with the event.
            local_port (varint): Local port associated with the event.
            remote_port (varint): Local port associated with the event.
            outbound (boolean): True in case of outbound traffic/connection.
            begin_time (datetime): Start of the event.
            end_time (datetime): End of the event.
            repetition (varint): How many times this event happened within the time frame.
            blocked (boolean): Whether the traffic/connection was succesfully blocked.
            severity (string): Severity of the event.
            rule_id (varint): Firewall rule ID associated with this event.
            rule_name (string): Name of the Firewall rule associated with this event.
            remote_host (string): Name of the remote host if it can be traced.
            application (path): Application responsible for/affected by event.
            user (string): User associated with the event.
            line_no (varint): Reference line number in log file.
        """

        for log_path in self.target.fs.glob(self.LOG_SEP_NET):
            log = self.target.fs.path(log_path).open("rb")
            line_no = 0
            while line := log.readline():
                line_no += 1
                if line_no == 1:
                    continue
                yield from self._fw_line(line, line_no)



from typing import Iterator

from dissect.target.helpers import keychain
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.keyprovider import (
    KeyProviderPlugin,
)


class KeychainKeyProviderPlugin(KeyProviderPlugin):
    """Keychain key provider plugin."""

    __namespace__ = "_dpapi_keyprovider_keychain"

    def check_compatible(self) -> None:
        return

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield keychain passphrases."""
        for key in keychain.get_keys_for_provider("user") + keychain.get_keys_without_provider():
            if key.key_type == keychain.KeyType.PASSPHRASE:
                yield self.__namespace__, key.value



from __future__ import annotations

import datetime
from typing import Iterator
from uuid import UUID

from dissect.cstruct import cstruct
from dissect.sql import sqlite3
from dissect.util.ts import wintimestamp
from flow.record import GroupedRecord

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import WindowsUserRecord, create_extended_descriptor
from dissect.target.plugin import Plugin, export

appdb_def = """
typedef struct {
    char     Magic[4];                 // Always b"DNPW"
    DWORD    Version;                  // Versions 1 (win8) and 3 (win10) seem to exist
    QWORD    Timestamp;                // According to some sources this is LastNotificationDate,
                                       // but that seems incorrect.
    DWORD    NextNotificationId;
    DWORD    Unknown;
    char     Padding[8];
} ChunkHeader;                         // size: 0x20

typedef struct {
    WORD     InUse;                    // ??
    WORD     NextTileWriteIndex;       // ??
    WORD     NextToastWriteIndex;      // ??
    BYTE     Flags[2];
} ChunkInfo;

typedef struct {
    QWORD    Timestamp1;               // ??
    QWORD    Timestamp2;               // Is this time to refresh?
    char     Uri[1024];                // Is this the correct size?
    char     Padding[0x818 - 0x410];
} PushDescriptor;                      // size: 0x818

typedef struct {
    DWORD    Id;
    DWORD    Zero;                     // ??
    QWORD    Timestamp;                // ??
    WORD     Unknown;
    WORD     DataLength;
    char     Data[DataLength];
    char     Padding[0x118 - 0x14 - DataLength];
} BadgeContent;                        // size: 0x118

typedef struct {
    DWORD    UniqueId;                 // ??
    DWORD    Zero;
    QWORD    ExpiryTime;               // The time this tile expires
    QWORD    ArrivalTime;              // The time this tile was set
    BYTE     Type;                     // ??
    BYTE     Index;
    WORD     ContentLength;
    wchar_t  Name[18];
} TileDescriptor;                      // size: 0x40

typedef struct {
    DWORD    UniqueId;                 // ??
    DWORD    Zero;
    QWORD    ExpiryTime;               // The time this toast expires
    QWORD    ArrivalTime;              // The time this toast was set
    BYTE     Type;                     // ??
    BYTE     Index;
    WORD     ContentLength;
    wchar_t  Name1[17];
    wchar_t  Name2[17];
} ToastDescriptor;                     // size: 0x60

typedef struct {
  char Content[0x1400];
} DataXML;                             // size: 0x1400

typedef struct {
    ChunkHeader     Header;            // Only populated for first chunk, else zeroed
    ChunkInfo       Info;
    PushDescriptor  Push;
    BadgeContent    Badge;
    TileDescriptor  Tiles[5];          // start @ 0x958
    DataXML         TileXml[5];

    // For the in use chunks, 0x14 ToastDiscriptors have an Index, but there
    // is space for more. Maybe this is used in case of deleted entries?
    ToastDescriptor Toasts[0x14];      // start @ 0x6e98
    char            Padding1[0x1e00];  // start @ 0x7618
    DataXML         ToastXml[0x14];    // start @ 0x9418
    char            Padding2[0x13f8];  // start @ 0x22418
} Chunk;                               // size: 0x23810
"""

c_appdb = cstruct(endian="<").load(appdb_def)

APPDB_MAGIC = b"DNPW"
NUM_APPDB_CHUNKS = 256

AppDBRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb",
    [
        ("datetime", "timestamp"),
        ("varint", "version"),
        ("varint", "next_notification_id"),
    ],
)

AppDBPushRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/push",
    [
        ("datetime", "push_ts1"),
        ("datetime", "push_ts2"),
        ("varint", "chunk_num"),
        ("uri", "push_uri"),
    ],
)

AppDBBadgeRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/badge",
    [
        ("datetime", "badge_ts"),
        ("varint", "badge_id"),
        ("varint", "chunk_num"),
        ("string", "badge_data"),
    ],
)

AppDBTileRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/tile",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "id"),
        ("varint", "chunk_num"),
        ("varint", "type"),
        ("varint", "index"),
        ("string", "name"),
        ("string", "content"),
    ],
)

AppDBToastRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/toast",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "id"),
        ("varint", "chunk_num"),
        ("varint", "type"),
        ("varint", "index"),
        ("string", "name1"),
        ("string", "name2"),
        ("string", "content"),
    ],
)

WpnDatabaseNotificationRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/wpndatabase",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "order"),
        ("varint", "id"),
        ("varint", "handler_id"),
        ("string", "activity_id"),
        ("string", "type"),
        ("bytes", "payload"),
        ("string", "payload_type"),
        ("string", "tag"),
        ("string", "group"),
        ("varint", "boot_id"),
        ("boolean", "expires_on_reboot"),
    ],
)

WpnDatabaseNotificationHandlerRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/wpndatabase/handler",
    [
        ("datetime", "created_time"),
        ("datetime", "modified_time"),
        ("varint", "id"),
        ("string", "primary_id"),
        ("string", "wns_id"),
        ("string", "handler_type"),
        ("varint", "wnf_event_name"),
        ("bytes", "system_data_property_set"),
    ],
)

NOTIFICATIONS_DIR = "AppData/Local/Microsoft/Windows/Notifications"


class NotificationsPlugin(Plugin):
    """Plugin that parses the notification databases on Windows 10 machines."""

    __namespace__ = "notifications"

    def __init__(self, target):
        super().__init__(target)
        self.wpndb_files = []
        self.appdb_files = []

        for user_details in target.user_details.all_with_home():
            notification_dir = user_details.home_path.joinpath(NOTIFICATIONS_DIR)
            wpndb_file = notification_dir.joinpath("wpndatabase.db")
            appdb_file = notification_dir.joinpath("appdb.dat")

            if wpndb_file.exists():
                self.wpndb_files.append((user_details.user, wpndb_file))

            if appdb_file.exists():
                with appdb_file.open(mode="rb") as fp:
                    chunk = c_appdb.Chunk(fp)
                    if chunk.Header.Magic == APPDB_MAGIC:
                        version = chunk.Header.Version
                        if version == 3:
                            self.appdb_files.append((user_details.user, appdb_file))
                        else:
                            self.target.log.warning(
                                "Skipping %s: unsupported version %s.",
                                appdb_file,
                                version,
                            )
                            if version != 1:
                                self.target.log.warning(
                                    "Unknown appdb version %s in file %s, "
                                    "please consider providing us with a sample.",
                                    version,
                                    appdb_file,
                                )

    def check_compatible(self) -> None:
        if not self.wpndb_files and not self.appdb_files:
            raise UnsupportedPluginError("No or incompatible wpndatabase.db or appdb.dat files found")

    def _get_appdb_chunk_record(
        self,
        chunk: c_appdb.Chunk,
        user: WindowsUserRecord,
    ) -> AppDBRecord:
        chunk_timestamp = None
        if ts := chunk.Header.Timestamp:
            chunk_timestamp = wintimestamp(ts)

        return AppDBRecord(
            timestamp=chunk_timestamp,
            version=chunk.Header.Version,
            next_notification_id=chunk.Header.NextNotificationId,
        )

    def _get_appdb_push_record(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> AppDBPushRecord | None:
        badge_record = None
        push_uri = chunk.Push.Uri.split(b"\x00")[0]
        push_uri = push_uri.decode("utf-8", errors="surrogateescape")

        if push_uri:
            push_timestamp1 = None
            if ts := chunk.Push.Timestamp1:
                push_timestamp1 = wintimestamp(ts)
            push_timestamp2 = None
            if ts := chunk.Push.Timestamp2:
                push_timestamp2 = wintimestamp(ts)

            badge_record = AppDBPushRecord(
                push_ts1=push_timestamp1,
                push_ts2=push_timestamp2,
                chunk_num=chunk_num,
                push_uri=push_uri,
                _target=self.target,
                _user=user,
            )

        return badge_record

    def _get_appdb_badge_record(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> AppDBBadgeRecord | None:
        badge_record = None
        badge_id = chunk.Badge.Id

        if badge_id:
            badge_ts = None
            if ts := chunk.Badge.Timestamp:
                badge_ts = wintimestamp(ts)
            badge_data = chunk.Badge.Data.decode("utf-8", errors="surrogateescape")

            badge_record = AppDBBadgeRecord(
                badge_id=badge_id,
                badge_ts=badge_ts,
                chunk_num=chunk_num,
                badge_data=badge_data,
                _target=self.target,
                _user=user,
            )

        return badge_record

    def _get_appdb_tile_records(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> list[AppDBTileRecord]:
        tile_records = []
        num_tiles = len(chunk.Tiles)

        for tile_no in range(num_tiles):
            tile = chunk.Tiles[tile_no]

            if tile.UniqueId:
                tile_arrival_time = None
                if ts := tile.ArrivalTime:
                    tile_arrival_time = wintimestamp(ts)
                tile_expiry_time = None
                if ts := tile.ExpiryTime:
                    tile_expiry_time = wintimestamp(ts)
                name = tile.Name.strip("\x00")

                xml_size = tile.ContentLength
                tile_xml = chunk.TileXml[tile_no].Content[:xml_size]
                tile_xml = tile_xml.decode("utf-8", errors="surrogateescape")

                tile_record = AppDBTileRecord(
                    arrival_time=tile_arrival_time,
                    expiry_time=tile_expiry_time,
                    id=tile.UniqueId,
                    chunk_num=chunk_num,
                    type=tile.Type,
                    index=tile.Index,
                    name=name,
                    content=tile_xml,
                    _target=self.target,
                    _user=user,
                )

                tile_records.append(tile_record)
        return tile_records

    def _get_appdb_toast_records(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> list[AppDBToastRecord]:
        toast_records = []
        num_toasts = len(chunk.Toasts)

        for toast_no in range(num_toasts):
            toast = chunk.Toasts[toast_no]

            if toast.UniqueId:
                toast_arrival_time = None
                if ts := toast.ArrivalTime:
                    toast_arrival_time = wintimestamp(ts)
                toast_expiry_time = None
                if ts := toast.ExpiryTime:
                    toast_expiry_time = wintimestamp(ts)
                name1 = toast.Name1.strip("\x00")
                name2 = toast.Name2.strip("\x00")

                xml_size = toast.ContentLength
                toast_xml = chunk.ToastXml[toast_no].Content[:xml_size]
                toast_xml = toast_xml.decode("utf-8", errors="surrogateescape")

                toast_record = AppDBToastRecord(
                    arrival_time=toast_arrival_time,
                    expiry_time=toast_expiry_time,
                    id=toast.UniqueId,
                    chunk_num=chunk_num,
                    type=toast.Type,
                    index=toast.Index,
                    name1=name1,
                    name2=name2,
                    content=toast_xml,
                    _target=self.target,
                    _user=user,
                )

                toast_records.append(toast_record)
        return toast_records

    @export(record=[AppDBRecord, AppDBPushRecord, AppDBBadgeRecord, AppDBTileRecord, AppDBToastRecord])
    def appdb(self) -> Iterator[GroupedRecord]:
        """Retrun the data from Windows appdb.dat file.

        This file contains data presentted to the user, pushed by external
        sources. The appdb.dat file was used from Windows 8 to Windows 10 pre
        anniversary version. This plugin only supports appdb.dat version 3 from
        Windows 10.

        References:
            - http://www.swiftforensics.com/2016/06/prasing-windows-10-notification-database.html
        """
        for user, appdb_file in self.appdb_files:
            with appdb_file.open(mode="rb") as fp:
                for chunk_num in range(NUM_APPDB_CHUNKS):
                    chunk = c_appdb.Chunk(fp)

                    if chunk.Info.InUse == 0:
                        continue
                    elif chunk.Info.InUse != 1:
                        self.target.log.warning(
                            "Unknown field value %s for chunk.Info.InUse, "
                            "please consider providing us with a sample.",
                            chunk.Info.InUse,
                        )
                        continue

                    if chunk_num == 0:
                        yield self._get_appdb_chunk_record(chunk, user)

                    push_record = self._get_appdb_push_record(chunk, chunk_num, user)
                    if push_record:
                        yield push_record

                    badge_record = self._get_appdb_badge_record(chunk, chunk_num, user)
                    if badge_record:
                        yield badge_record

                    for tile_record in self._get_appdb_tile_records(chunk, chunk_num, user):
                        yield tile_record

                    for toast_record in self._get_appdb_toast_records(chunk, chunk_num, user):
                        yield toast_record

    @export(record=[WpnDatabaseNotificationRecord, WpnDatabaseNotificationHandlerRecord])
    def wpndatabase(self) -> Iterator[WpnDatabaseNotificationRecord | WpnDatabaseNotificationHandlerRecord]:
        """Returns Windows Notifications from wpndatabase.db (post Windows 10 Anniversary).

        References:
            - https://inc0x0.com/2018/10/windows-10-notification-database/
        """
        for user, wpndatabase in self.wpndb_files:
            db = sqlite3.SQLite3(wpndatabase.open())

            handlers = {}
            for row in db.table("NotificationHandler").rows():
                handlers[row["[RecordId]"]] = WpnDatabaseNotificationHandlerRecord(
                    created_time=datetime.datetime.strptime(row["[CreatedTime]"], "%Y-%m-%d %H:%M:%S"),
                    modified_time=datetime.datetime.strptime(row["[ModifiedTime]"], "%Y-%m-%d %H:%M:%S"),
                    id=row["[RecordId]"],
                    primary_id=row["[PrimaryId]"],
                    wns_id=row["[WNSId]"],
                    handler_type=row["[HandlerType]"],
                    wnf_event_name=row["[WNFEventName]"],
                    system_data_property_set=row["[SystemDataPropertySet]"],
                    _target=self.target,
                    _user=user,
                )

            for row in db.table("Notification").rows():
                record = WpnDatabaseNotificationRecord(
                    arrival_time=wintimestamp(row["[ArrivalTime]"]),
                    expiry_time=wintimestamp(row["[ExpiryTime]"]),
                    order=row["[Order]"],
                    id=row["[Id]"],
                    handler_id=row["[HandlerId]"],
                    activity_id=UUID(bytes=row["[ActivityId]"]),
                    type=row["[Type]"],
                    payload=row["[Payload]"],
                    payload_type=row["[PayloadType]"],
                    tag=row["[Tag]"],
                    group=row["[Group]"],
                    boot_id=row["[BootId]"],
                    expires_on_reboot=row["[ExpiresOnReboot]"] != "FALSE",
                    _target=self.target,
                    _user=user,
                )
                handler = handlers.get(row["[HandlerId]"])

                if handler:
                    yield GroupedRecord("windows/notification/wpndatabase/grouped", [record, handler])
                else:
                    yield record



import argparse
import dataclasses
import textwrap
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set, Type

from dissect.target import Target
from dissect.target.plugin import Plugin
from dissect.target.target import Event

BLOCK_INDENT = 4 * " "


@dataclass
class TargetExecutionReport:
    target: Target

    incompatible_plugins: Set[str] = dataclasses.field(default_factory=set)
    registered_plugins: Set[str] = dataclasses.field(default_factory=set)

    func_errors: Dict[str, str] = dataclasses.field(default_factory=dict)
    func_execs: Set[str] = dataclasses.field(default_factory=set)

    def add_incompatible_plugin(self, plugin_name: str) -> None:
        self.incompatible_plugins.add(plugin_name)

    def add_registered_plugin(self, plugin_name: str) -> None:
        self.registered_plugins.add(plugin_name)

    def add_func_error(self, func, stacktrace: str) -> None:
        self.func_errors[func] = stacktrace

    def as_dict(self) -> Dict[str, Any]:
        return {
            "target": str(self.target),
            "incompatible_plugins": sorted(self.incompatible_plugins),
            "registered_plugins": sorted(self.registered_plugins),
            "func_errors": self.func_errors,
            "func_execs": sorted(self.func_execs),
        }


@dataclass
class ExecutionReport:
    plugin_import_errors: Dict[str, str] = dataclasses.field(default_factory=dict)

    target_reports: List[TargetExecutionReport] = dataclasses.field(default_factory=list)

    cli_args: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def set_cli_args(self, args: argparse.Namespace) -> None:
        args = ((key, str(value)) for (key, value) in vars(args).items())
        self.cli_args.update(args)

    def set_plugin_stats(self, plugins: Dict[str, Any]) -> None:
        for details in plugins.get("_failed", []):
            self.plugin_import_errors[details["module"]] = "".join(details["stacktrace"])

    def get_formatted_report(self) -> str:
        blocks = [
            make_cli_args_overview(self),
            make_plugin_import_errors_overview(self),
            *[format_target_report(report) for report in self.target_reports],
        ]
        return "\n".join(blocks)

    def add_target_report(self, target: Target) -> TargetExecutionReport:
        target_report = TargetExecutionReport(target=target)
        self.target_reports.append(target_report)
        return target_report

    def get_target_report(self, target: Target, create: bool = False) -> TargetExecutionReport:
        target_report = next(filter(lambda r: r.target == target, self.target_reports), None)
        if target_report is None and create:
            target_report = self.add_target_report(target)
        return target_report

    @staticmethod
    def _get_plugin_name(plugin_cls):
        return f"{plugin_cls.__module__}.{plugin_cls.__qualname__}"

    def log_incompatible_plugin(
        self,
        target: Target,
        _,
        plugin_cls: Optional[Type[Plugin]] = None,
        plugin_desc: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not plugin_cls and not plugin_desc:
            raise ValueError("Either `plugin_cls` or `plugin_desc` must be set")

        target_report = self.get_target_report(target, create=True)

        if plugin_cls:
            plugin_name = self._get_plugin_name(plugin_cls)
        elif plugin_desc:
            plugin_name = plugin_desc["fullname"]

        target_report.add_incompatible_plugin(plugin_name)

    def log_registered_plugin(self, target: Target, _, plugin_inst: Plugin) -> None:
        target_report = self.get_target_report(target, create=True)
        plugin_cls = type(plugin_inst)
        plugin_name = self._get_plugin_name(plugin_cls)
        target_report.add_registered_plugin(plugin_name)

    def log_func_error(self, target: Target, _, func: str, stacktrace: str) -> None:
        target_report = self.get_target_report(target, create=True)
        target_report.add_func_error(func, stacktrace)

    def log_func_execution(self, target: Target, _, func: str) -> None:
        target_report = self.get_target_report(target, create=True)
        target_report.func_execs.add(func)

    def set_event_callbacks(self, target_cls: Type[Target]) -> None:
        target_cls.set_event_callback(
            event_type=Event.INCOMPATIBLE_PLUGIN,
            event_callback=self.log_incompatible_plugin,
        )
        target_cls.set_event_callback(
            event_type=Event.REGISTERED_PLUGIN,
            event_callback=self.log_registered_plugin,
        )
        target_cls.set_event_callback(
            event_type=Event.FUNC_EXEC,
            event_callback=self.log_func_execution,
        )
        target_cls.set_event_callback(
            event_type=Event.FUNC_EXEC_ERROR,
            event_callback=self.log_func_error,
        )

    def as_dict(self) -> Dict[str, Any]:
        return {
            "plugin_import_errors": self.plugin_import_errors,
            "target_reports": [report.as_dict() for report in self.target_reports],
            "cli_args": self.cli_args,
        }


def make_cli_args_overview(report: ExecutionReport) -> str:
    header = "CLI arguments:"

    rows = []
    for key, value in sorted(report.cli_args.items()):
        rows.append(f"{key}: {value}")

    block = "\n".join(rows)
    block = textwrap.indent(block, prefix=BLOCK_INDENT)
    return "\n".join([header, block])


def make_plugin_import_errors_overview(report: ExecutionReport, short=True) -> str:
    header = "Plugin import errors:"

    rows = []
    for module, trace in sorted(report.plugin_import_errors.items()):
        trace = trace.rstrip()
        trace = textwrap.indent(trace, prefix=BLOCK_INDENT)
        rows.append(f"{module}:\n{trace}")

    block = "\n".join(rows)
    block = textwrap.indent(block, prefix=BLOCK_INDENT)
    return "\n".join([header, block])


def format_target_report(target_report: TargetExecutionReport) -> str:
    blocks = [f"Target: {target_report.target}"]

    registered_plugins_header = "Registered plugins:"
    registered_plugins_rows = sorted(target_report.registered_plugins)
    registered_plugins_rows_block = textwrap.indent(
        "\n".join(registered_plugins_rows),
        prefix=BLOCK_INDENT,
    )
    registered_plugins_block = textwrap.indent(
        "\n".join([registered_plugins_header, registered_plugins_rows_block]),
        prefix=BLOCK_INDENT,
    )
    blocks.append(registered_plugins_block)

    incompatible_plugins_header = "Incompatible plugins:"
    incompatible_plugins_rows = sorted(target_report.incompatible_plugins)
    incompatible_plugins_rows_block = textwrap.indent(
        "\n".join(incompatible_plugins_rows),
        prefix=BLOCK_INDENT,
    )
    incompatible_plugins_block = textwrap.indent(
        "\n".join([incompatible_plugins_header, incompatible_plugins_rows_block]),
        prefix=BLOCK_INDENT,
    )
    blocks.append(incompatible_plugins_block)

    func_errors_header = "Function errors:"
    func_errors_rows = []
    for func, stacktrace in sorted(target_report.func_errors.items()):
        stacktrace = textwrap.indent(stacktrace, prefix=BLOCK_INDENT)
        func_errors_rows.append(f"{func}:\n{stacktrace}")

    func_errors_rows_block = textwrap.indent("\n".join(func_errors_rows), prefix=BLOCK_INDENT)
    func_errors_block = textwrap.indent(
        "\n".join([func_errors_header, func_errors_rows_block]),
        prefix=BLOCK_INDENT,
    )
    blocks.append(func_errors_block)

    func_execs_header = "Function executions:"
    func_execs_rows = sorted(target_report.func_execs)
    func_execs_rows_block = textwrap.indent("\n".join(func_execs_rows), prefix=BLOCK_INDENT)
    func_execs_block = textwrap.indent(
        "\n".join([func_execs_header, func_execs_rows_block]),
        prefix=BLOCK_INDENT,
    )
    blocks.append(func_execs_block)

    return "\n".join(blocks)



from unittest.mock import Mock, patch

import dissect.target.plugins.general.plugins as plugin
from dissect.target.plugins.general.plugins import (
    PluginListPlugin,
    categorize_plugins,
    dictify_module_recursive,
    output_plugin_description_recursive,
    update_dict_recursive,
)


def test_dictify_module():
    last_value = Mock()

    output_dict = dictify_module_recursive(["hello", "world"], last_value)

    assert output_dict == {"hello": {"world": last_value}}


def test_update_dict():
    tmp_dictionary = dict()

    update_dict_recursive(tmp_dictionary, dictify_module_recursive(["hello", "world"], None))
    update_dict_recursive(tmp_dictionary, dictify_module_recursive(["hello", "lawrence"], None))

    assert tmp_dictionary == {"hello": {"world": None, "lawrence": None}}


def test_plugin_description():
    description = [x for x in output_plugin_description_recursive(PluginListPlugin, False)]
    assert description == ["plugins - Print all available plugins. (output: no output)"]


def test_plugin_description_compacting():
    module = dictify_module_recursive(["hello", "world"], PluginListPlugin)

    description = [x for x in output_plugin_description_recursive(module, False)]
    assert description == [
        "hello:",
        "  world:",
        "    plugins - Print all available plugins. (output: no output)",
    ]


def test_plugin_description_in_dict_multiple():
    tmp_dictionary = dict()

    update_dict_recursive(tmp_dictionary, dictify_module_recursive(["hello", "world", "data"], PluginListPlugin))
    update_dict_recursive(tmp_dictionary, dictify_module_recursive(["hello", "world", "data2"], PluginListPlugin))

    description = [x for x in output_plugin_description_recursive(tmp_dictionary, False)]
    assert description == [
        "hello:",
        "  world:",
        "    data:",
        "      plugins - Print all available plugins. (output: no output)",
        "    data2:",
        "      plugins - Print all available plugins. (output: no output)",
    ]


@patch.object(plugin.plugin, "load")
@patch.object(plugin, "get_exported_plugins")
def test_categorize_plugins(mocked_export, mocked_load):
    mocked_export.return_value = [{"module": "something.data"}]
    assert categorize_plugins() == {"something": {"data": mocked_load.return_value}}



from datetime import datetime, timezone, tzinfo
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from dissect.target.plugin import Plugin, internal
from dissect.target.target import Target


class DateTimePlugin(Plugin):
    __namespace__ = "datetime"

    def __init__(self, target: Target):
        super().__init__(target)
        self.timezone = target.timezone
        if not self.timezone:
            self.target.log.warning("Could not determine timezone of target, falling back to UTC for datetime helpers.")

    def check_compatible(self) -> None:
        pass

    @internal
    def tz(self, name: str) -> tzinfo:
        return ZoneInfo(name)

    @internal(property=True)
    def tzinfo(self) -> tzinfo:
        """Return a datetime.tzinfo of the current system timezone."""
        try:
            return self.tz(self.timezone)
        except (TypeError, ZoneInfoNotFoundError):
            return timezone.utc

    @internal
    def local(self, dt: datetime) -> datetime:
        """Replace the tzinfo of a given datetime.datetime object with the current system tzinfo without conversion."""
        return dt.replace(tzinfo=self.tzinfo)

    @internal
    def to_utc(self, dt: datetime) -> datetime:
        """Convert any datetime.datetime object into a UTC datetime.datetime object.

        First replaces the current tzinfo with the system tzinfo without conversion, then converts it to an aware
        UTC datetime object.
        """
        return self.local(dt).astimezone(timezone.utc)



from dissect.target import Target
from dissect.target.plugins.filesystem.acquire_handles import OpenHandlesPlugin
from tests._utils import absolute_path


def test_acquire_handles_plugin():
    file_hashes_target = Target().open(
        absolute_path("_data/plugins/filesystem/acquire_handles/test-acquire-handles.tar")
    )
    file_hashes_target.add_plugin(OpenHandlesPlugin)

    results = list(file_hashes_target.acquire_handles())
    first_result = results[0]

    assert first_result.name == r"\Windows\Fonts"
    assert first_result.handle_type == "EtwRegistration"
    assert first_result.unique_process_id == 1
    assert first_result.object == "0xfffftest"
    assert results[-1].unique_process_id == 124
    assert len(results) == 124



import base64
import re
from itertools import product
from pathlib import Path
from typing import Iterator

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugin import export
from dissect.target.plugins.apps.ssh.ssh import (
    AuthorizedKeysRecord,
    KnownHostRecord,
    PrivateKeyRecord,
    PublicKeyRecord,
    SSHPlugin,
    SSHPrivateKey,
    calculate_fingerprints,
)


def find_sshd_directory(target: Target) -> TargetPath:
    SSHD_DIRECTORIES = ["/sysvol/ProgramData/ssh", "/etc/ssh"]

    for sshd in SSHD_DIRECTORIES:
        if (target_path := target.fs.path(sshd)).exists():
            return target_path

    # A default, so there is no need to check for None
    return target.fs.path("/etc/ssh/")


class OpenSSHPlugin(SSHPlugin):
    """OpenSSH plugin."""

    __namespace__ = "openssh"

    SSHD_DIRECTORIES = ["/sysvol/ProgramData/ssh", "/etc/ssh"]

    def __init__(self, target: Target):
        super().__init__(target)
        self.sshd_directory = find_sshd_directory(target)

    def check_compatible(self) -> None:
        ssh_user_dirs = any(
            user_details.home_path.joinpath(".ssh").exists()
            for user_details in self.target.user_details.all_with_home()
        )
        if not ssh_user_dirs and not self.sshd_directory.exists():
            raise UnsupportedPluginError("No OpenSSH directories found")

    def ssh_directory_globs(self, glob_user: str, glob_sshd: str) -> list[tuple[str, TargetPath]]:
        for user_details in self.target.user_details.all_with_home():
            yield from product([user_details.user], user_details.home_path.glob(f".ssh/{glob_user}"))

        yield from product([None], self.sshd_directory.glob(glob_sshd))

    @export(record=AuthorizedKeysRecord)
    def authorized_keys(self) -> Iterator[AuthorizedKeysRecord]:
        """Yields the content of the authorized_keys files on a target for each user."""
        for user, authorized_keys_file in self.ssh_directory_globs("authorized_keys*", "administrator_authorized_keys"):
            for line in authorized_keys_file.open("rt"):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                try:
                    options, keytype, public_key, comment = parse_ssh_key(line)
                except ValueError:
                    continue

                yield AuthorizedKeysRecord(
                    key_type=keytype,
                    public_key=public_key,
                    comment=comment,
                    options=options,
                    path=authorized_keys_file,
                    _target=self.target,
                    _user=user,
                )

    @export(record=KnownHostRecord)
    def known_hosts(self) -> Iterator[KnownHostRecord]:
        """Yields the content of the known_hosts files on a target for each user."""

        for user, known_hosts_file in self.ssh_directory_globs("known_hosts*", "ssh_known_hosts"):
            for line in known_hosts_file.open("rt"):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                try:
                    marker, hostnames, keytype, public_key, comment = parse_known_host(line)
                except ValueError:
                    continue

                for hostname in hostnames:
                    yield KnownHostRecord(
                        host=hostname,
                        port=None,
                        key_type=keytype,
                        public_key=public_key,
                        comment=comment,
                        marker=marker,
                        path=known_hosts_file,
                        _target=self.target,
                        _user=user,
                    )

    @export(record=PrivateKeyRecord)
    def private_keys(self) -> Iterator[PrivateKeyRecord]:
        """Yields OpenSSH private keys on a target for each user."""

        for user, file_path in self.ssh_directory_globs("*", "*"):
            if not file_path.is_file():
                continue

            buffer = file_path.read_bytes().strip()
            if b"PRIVATE KEY-----" not in buffer:
                continue

            try:
                private_key = SSHPrivateKey(buffer)
            except ValueError as e:
                self.target.log.warning("Failed to parse SSH private key: %s", file_path)
                self.target.log.debug("", exc_info=e)
                continue

            yield PrivateKeyRecord(
                mtime_ts=file_path.stat().st_mtime,
                key_format=private_key.format,
                key_type=private_key.key_type,
                public_key=private_key.public_key,
                comment=private_key.comment,
                encrypted=private_key.is_encrypted,
                path=file_path,
                _target=self.target,
                _user=user,
            )

    @export(record=PublicKeyRecord)
    def public_keys(self) -> Iterator[PublicKeyRecord]:
        """Yields all OpenSSH public keys from all user home directories and the OpenSSH daemon directory."""

        for user, file_path in self.ssh_directory_globs("*.pub", "*.pub"):
            if not file_path.is_file():
                continue

            key_type, public_key, comment = parse_ssh_public_key_file(file_path)
            fingerprints = calculate_fingerprints(base64.b64decode(public_key))

            yield PublicKeyRecord(
                mtime_ts=file_path.stat().st_mtime,
                key_type=key_type,
                public_key=public_key,
                comment=comment,
                fingerprint=fingerprints,
                path=file_path,
                _target=self.target,
                _user=user,
            )


def parse_ssh_public_key_file(path: Path) -> tuple[str, str, str]:
    _, key_type, public_key, comment = parse_ssh_key(path.read_text().strip())
    return key_type, public_key, comment


def parse_ssh_key(key_string: str) -> tuple[str, str, str, str]:
    parts = re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', key_string)

    options = None
    if not key_string.startswith(("sk-", "ssh-", "ecdsa-")):
        options = parts.pop(0)

    keytype, public_key = parts[:2]

    # We strip whitespace to make parsing a bit safer, but this means we also strip "empty" comments
    # An empty comment is technically "", so return that instead of None
    comment = " ".join(parts[2:]) if len(parts) > 2 else ""

    return options, keytype, public_key, comment


def parse_known_host(known_host_string: str) -> tuple[str, list, str, str, str]:
    parts = known_host_string.split()

    marker = None
    if parts[0].startswith("@"):
        marker = parts.pop(0)

    hostnames, keytype, public_key = parts[:3]

    # We strip whitespace to make parsing a bit safer, but this means we also strip "empty" comments
    # An empty comment is technically "", so return that instead of None
    comment = " ".join(parts[3:]) if len(parts) > 3 else ""

    return marker, hostnames.split(","), keytype, public_key, comment



from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import UnixApplicationRecord
from dissect.target.plugin import Plugin, export
from dissect.target.target import Target


class UnixApplicationsPlugin(Plugin):
    """Unix Applications plugin."""

    SYSTEM_PATHS = [
        "/usr/share/applications/",
        "/usr/local/share/applications/",
        "/var/lib/snapd/desktop/applications/",
        "/var/lib/flatpak/exports/share/applications/",
    ]

    USER_PATHS = [
        ".local/share/applications/",
    ]

    SYSTEM_APPS = ("org.gnome.",)

    def __init__(self, target: Target):
        super().__init__(target)
        self.desktop_files = list(self._find_desktop_files())

    def _find_desktop_files(self) -> Iterator[TargetPath]:
        for dir in self.SYSTEM_PATHS:
            for file in self.target.fs.path(dir).glob("*.desktop"):
                yield file

        for user_details in self.target.user_details.all_with_home():
            for dir in self.USER_PATHS:
                for file in user_details.home_path.joinpath(dir).glob("*.desktop"):
                    yield file

    def check_compatible(self) -> None:
        if not self.desktop_files:
            raise UnsupportedPluginError("No application .desktop files found")

    @export(record=UnixApplicationRecord)
    def applications(self) -> Iterator[UnixApplicationRecord]:
        """Yield installed Unix GUI applications from GNOME and XFCE.

        Resources:
            - https://wiki.archlinux.org/title/Desktop_entries
            - https://specifications.freedesktop.org/desktop-entry-spec/latest/
            - https://unix.stackexchange.com/questions/582928/where-gnome-apps-are-installed

        Yields ``UnixApplicationRecord`` records with the following fields:

        .. code-block:: text

            ts_modified  (datetime): timestamp when the installation was modified
            ts_installed (datetime): timestamp when the application was installed on the system
            name         (string):   name of the application
            version      (string):   version of the application
            author       (string):   author of the application
            type         (string):   type of the application, either user or system
            path         (string):   path to the desktop file entry of the application
        """
        for file in self.desktop_files:
            config = configutil.parse(file, hint="ini").get("Desktop Entry") or {}
            stat = file.lstat()

            yield UnixApplicationRecord(
                ts_modified=stat.st_mtime,
                ts_installed=stat.st_btime if hasattr(stat, "st_btime") else None,
                name=config.get("Name"),
                version=config.get("Version"),
                path=config.get("Exec"),
                type="system" if config.get("Icon", "").startswith(self.SYSTEM_APPS) else "user",
                _target=self.target,
            )



from __future__ import annotations

import logging
import re
import tarfile
from pathlib import Path

from dissect.target import filesystem, target
from dissect.target.filesystems.tar import (
    TarFilesystemDirectoryEntry,
    TarFilesystemEntry,
)
from dissect.target.helpers import fsutil, loaderutil
from dissect.target.loader import Loader

log = logging.getLogger(__name__)


ANON_FS_RE = re.compile(r"^fs[0-9]+$")


class TarLoader(Loader):
    """Load tar files."""

    def __init__(self, path: Path | str, **kwargs):
        super().__init__(path)

        if isinstance(path, str):
            path = Path(path)

        if self.is_compressed(path):
            log.warning(
                f"Tar file {path!r} is compressed, which will affect performance. "
                "Consider uncompressing the archive before passing the tar file to Dissect."
            )

        self.tar = tarfile.open(fileobj=path.open("rb"))

    @staticmethod
    def detect(path: Path) -> bool:
        return path.name.lower().endswith((".tar", ".tar.gz", ".tgz"))

    def is_compressed(self, path: Path | str) -> bool:
        return str(path).lower().endswith((".tar.gz", ".tgz"))

    def map(self, target: target.Target) -> None:
        volumes = {}

        for member in self.tar.getmembers():
            if member.name == ".":
                continue

            if not member.name.startswith(("/fs/", "fs/", "/sysvol/", "sysvol/")):
                # Not an acquire tar
                if "/" not in volumes:
                    vol = filesystem.VirtualFilesystem(case_sensitive=True)
                    vol.tar = self.tar
                    volumes["/"] = vol
                    target.filesystems.add(vol)

                volume = volumes["/"]
                mname = member.name
            else:
                if member.name.startswith(("/fs/", "fs/")):
                    # Current acquire
                    parts = member.name.replace("fs/", "").split("/")
                    if parts[0] == "":
                        parts.pop(0)
                else:
                    # Legacy acquire
                    parts = member.name.lstrip("/").split("/")
                volume_name = parts[0].lower()

                # NOTE: older versions of acquire would write to "sysvol" instead of a driver letter
                # Figuring out the sysvol from the drive letters is easier than the drive letter from "sysvol",
                # so this was swapped in acquire 3.12. Now we map all volumes to a drive letter and let the
                # Windows OS plugin figure out which is the sysvol
                # For backwards compatibility we're forced to keep this check, and assume that "c:" is our sysvol
                if volume_name == "sysvol":
                    volume_name = "c:"

                if volume_name == "$fs$":
                    if len(parts) == 1:
                        # The fs/$fs$ entry is ignored, only the directories below it are processed.
                        continue
                    fs_name = parts[1]
                    if ANON_FS_RE.match(fs_name):
                        parts.pop(0)
                        volume_name = f"{volume_name}/{fs_name}"

                if volume_name not in volumes:
                    vol = filesystem.VirtualFilesystem(case_sensitive=False)
                    vol.tar = self.tar
                    volumes[volume_name] = vol
                    target.filesystems.add(vol)

                volume = volumes[volume_name]
                mname = "/".join(parts[1:])

            entry_cls = TarFilesystemDirectoryEntry if member.isdir() else TarFilesystemEntry
            entry = entry_cls(volume, fsutil.normpath(mname), member)
            volume.map_file_entry(entry.path, entry)

        for vol_name, vol in volumes.items():
            loaderutil.add_virtual_ntfs_filesystem(
                target,
                vol,
                usnjrnl_path=[
                    "$Extend/$Usnjrnl:$J",
                    "$Extend/$Usnjrnl:J",  # Old versions of acquire used $Usnjrnl:J
                ],
            )

            target.fs.mount(vol_name, vol)



from datetime import datetime, timezone
from os import stat

from dissect.target.plugins.os.unix.generic import GenericPlugin
from tests._utils import absolute_path

log_mtime = stat(absolute_path("_data/plugins/os/unix/log/empty.log")).st_mtime


def test_unix_generic_activity(target_unix, fs_unix):
    fs_unix.map_file("/var/log/some.log", absolute_path("_data/plugins/os/unix/log/empty.log"))
    target_unix.add_plugin(GenericPlugin)
    assert target_unix.activity == datetime.fromtimestamp(log_mtime, tz=timezone.utc)


def test_unix_generic_install_date(target_unix, fs_unix):
    fs_unix.map_file("/var/log/installer/syslog", absolute_path("_data/plugins/os/unix/log/empty.log"))
    target_unix.add_plugin(GenericPlugin)
    assert target_unix.install_date == datetime.fromtimestamp(log_mtime, tz=timezone.utc)



import logging
from io import BytesIO
from unittest.mock import patch

import pytest
from pytest import LogCaptureFixture

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.registry import RegistryPlugin
from dissect.target.target import Target


def test_missing_hives(fs_win: VirtualFilesystem, caplog: LogCaptureFixture) -> None:
    target = Target()
    target.filesystems.add(fs_win)

    caplog.set_level(logging.DEBUG, target.log.name)
    target.apply()

    expected = []

    base_paths = [
        "sysvol/windows/system32/config",
        "sysvol/WINNT/system32/config",
        "sysvol/windows",
        "sysvol/reactos",
        "sysvol/windows/system32/config/RegBack",
    ]

    for base_path in base_paths:
        expected += [f"{target}: Could not find hive: {base_path}/{hive}" for hive in RegistryPlugin.SYSTEM]

    assert [record.message for record in caplog.records if record.filename == "registry.py"] == expected


def test_missing_user_hives(fs_win: VirtualFilesystem, target_win_users: Target, caplog: LogCaptureFixture) -> None:
    fs_win.makedirs("Users/John")

    caplog.set_level(logging.DEBUG, target_win_users.log.name)
    target_win_users.registry.load_user_hives()

    assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
        f"{target_win_users}: Could not find ntuser.dat: C:/Users/John/ntuser.dat",
        f"{target_win_users}: Could not find usrclass.dat: C:/Users/John/AppData/Local/Microsoft/Windows/usrclass.dat",
    ]


def test_empty_hives(fs_win: VirtualFilesystem, caplog: LogCaptureFixture) -> None:
    fs_win.map_file_fh("windows/system32/config/SYSTEM", BytesIO())
    fs_win.map_file_fh("boot/BCD", BytesIO())

    target = Target()
    target.filesystems.add(fs_win)

    caplog.set_level(logging.WARNING, target.log.name)
    target.apply()

    assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
        f"{target}: Empty hive: sysvol/windows/system32/config/SYSTEM",
        f"{target}: Empty BCD hive: sysvol/boot/BCD",
    ]


def test_empty_hives_skip_warning(fs_win: VirtualFilesystem, caplog: LogCaptureFixture) -> None:
    fake_fh = BytesIO()
    fake_fh.size = 1
    fs_win.map_file_fh("windows/system32/config/SYSTEM", fake_fh)
    fs_win.map_file_fh("windows/system32/config/RegBack/SYSTEM", BytesIO())

    target = Target()
    target.filesystems.add(fs_win)

    caplog.set_level(logging.WARNING, target.log.name)

    with patch("dissect.target.plugins.os.windows.registry.RegfHive"):
        target.apply()

    assert [record.message for record in caplog.records if record.filename == "registry.py"] == []


def test_empty_user_hives(fs_win: VirtualFilesystem, target_win_users: Target, caplog: LogCaptureFixture) -> None:
    fs_win.map_file_fh("Users/John/ntuser.dat", BytesIO())
    fs_win.map_file_fh("Users/John/AppData/Local/Microsoft/Windows/usrclass.dat", BytesIO())

    caplog.set_level(logging.WARNING, target_win_users.log.name)
    target_win_users.registry.load_user_hives()

    assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
        f"{target_win_users}: Empty NTUSER.DAT hive: C:/Users/John/ntuser.dat",
        f"{target_win_users}: Empty UsrClass.DAT hive: C:/Users/John/AppData/Local/Microsoft/Windows/usrclass.dat",
    ]


@pytest.mark.parametrize(
    "pattern, key_names",
    [
        (
            "\\HKLM\\SOFTWARE\\Microsoft",
            ["Microsoft"],
        ),
        (
            "\\*\\SOFTWARE\\Microsoft",
            ["Microsoft"],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList\\*\\",
            [
                "S-1-5-18",
                "S-1-5-21-3263113198-3007035898-945866154-1002",
            ],
        ),
        (
            "\\HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows XP\\*",
            [],
        ),
        (
            "\\*",
            [
                "HKEY_LOCAL_MACHINE",
                "HKEY_USERS",
            ],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\*\\CurrentVersion\\ProfileList",
            ["ProfileList"],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\*\\Windows NT\\*\\ProfileList\\",
            ["ProfileList"],
        ),
    ],
)
def test_registry_plugin_glob_ext(target_win_users: Target, pattern: str, key_names: list[str]) -> None:
    registry_plugin = target_win_users.registry

    key_collections = registry_plugin.glob_ext(pattern)
    collection_names = []
    for key_collection in key_collections:
        collection_names.append(key_collection.name)

    assert sorted(collection_names) == sorted(key_names)


def test_registry_plugin_root_none(target_win_users: Target) -> None:
    plugin: RegistryPlugin = target_win_users.registry

    assert plugin.key()
    assert plugin.key("")



from __future__ import annotations

import logging
from typing import Callable, Iterator

from dissect.ntfs.attr import Attribute
from dissect.ntfs.c_ntfs import FILE_RECORD_SEGMENT_IN_USE
from dissect.ntfs.mft import MftRecord
from flow.record import Record
from flow.record.fieldtypes import windows_path

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.filesystem.ntfs.utils import (
    InformationType,
    get_drive_letter,
    get_owner_and_group,
    get_record_size,
    get_volume_identifier,
)

log = logging.getLogger(__name__)


FilesystemStdCompactRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/std/compact",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_modification_time"),
        ("datetime", "last_change_time"),
        ("datetime", "last_access_time"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("string", "volume_uuid"),
    ],
)


FilesystemStdRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/std",
    [
        ("datetime", "ts"),
        ("string", "ts_type"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("string", "volume_uuid"),
    ],
)

FilesystemFilenameCompactRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/filename/compact",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_modification_time"),
        ("datetime", "last_change_time"),
        ("datetime", "last_access_time"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

FilesystemFilenameRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/filename",
    [
        ("datetime", "ts"),
        ("string", "ts_type"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

FilesystemMACBRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/macb",
    [
        ("datetime", "ts"),
        ("string", "macb"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

RECORD_TYPES = {
    InformationType.STANDARD_INFORMATION: FilesystemStdRecord,
    InformationType.FILE_INFORMATION: FilesystemFilenameRecord,
}


COMPACT_RECORD_TYPES = {
    InformationType.STANDARD_INFORMATION: FilesystemStdCompactRecord,
    InformationType.FILE_INFORMATION: FilesystemFilenameCompactRecord,
}


class MftPlugin(Plugin):
    """NTFS MFT plugin."""

    def __init__(self, target):
        super().__init__(target)
        self.ntfs_filesystems = {index: fs for index, fs in enumerate(self.target.filesystems) if fs.__type__ == "ntfs"}

    def check_compatible(self) -> None:
        if not len(self.ntfs_filesystems):
            raise UnsupportedPluginError("No NTFS filesystems found")

    @export(
        record=[
            FilesystemStdRecord,
            FilesystemFilenameRecord,
            FilesystemStdCompactRecord,
            FilesystemFilenameCompactRecord,
        ]
    )
    @arg(
        "--compact",
        group="fmt",
        action="store_true",
        help="compacts the MFT entry timestamps into a single record",
    )
    @arg("--fs", type=int, default=None, help="optional filesystem index, zero indexed")
    @arg("--start", type=int, default=0, help="the first MFT segment number")
    @arg("--end", type=int, default=-1, help="the last MFT segment number")
    @arg(
        "--macb",
        group="fmt",
        action="store_true",
        help="compacts MFT timestamps into MACB bitfield (format: MACB[standard|ads]/MACB[filename])",
    )
    def mft(
        self, compact: bool = False, fs: int | None = None, start: int = 0, end: int = -1, macb: bool = False
    ) -> Iterator[Record]:
        """Return the MFT records of all NTFS filesystems.

        The Master File Table (MFT) contains primarily metadata about every file and folder on a NFTS filesystem.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the MFT properties
        added to it through a "fake" ``NtfsFilesystem``), the paths returned in the MFT records are based on the
        mount point of the ``VirtualFilesystem``. This ensures that the proper original drive letter is used when
        available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/fileio/master-file-table
        """

        record_formatter = formatter

        def noaggr(records: list[Record]) -> Iterator[Record]:
            yield from records

        aggr = noaggr

        if compact:
            record_formatter = compacted_formatter
        elif macb:
            aggr = macb_aggr

        if fs is not None:
            try:
                filesystem = self.ntfs_filesystems[fs]
            except KeyError:
                self.target.log.error("NTFS filesystem with index number %s does not exist", fs)
                return

            yield from self.segments(filesystem, record_formatter, aggr, start, end)
        else:
            for filesystem in self.ntfs_filesystems.values():
                yield from self.segments(filesystem, record_formatter, aggr, start, end)

    def segments(
        self, fs: NtfsFilesystem, record_formatter: Callable, aggr: Callable, start: int, end: int
    ) -> Iterator[Record]:
        # If this filesystem is a "fake" NTFS filesystem, used to enhance a
        # VirtualFilesystem, The driveletter (more accurate mount point)
        # returned will be that of the VirtualFilesystem. This makes sure
        # the paths returned in the records are actually reachable.
        drive_letter = get_drive_letter(self.target, fs)
        volume_uuid = get_volume_identifier(fs)

        try:
            for record in fs.ntfs.mft.segments(start, end):
                try:
                    inuse = bool(record.header.Flags & FILE_RECORD_SEGMENT_IN_USE)
                    owner, _ = get_owner_and_group(record, fs)
                    resident = None
                    size = None

                    if not record.is_dir():
                        for data_attribute in record.attributes.DATA:
                            if data_attribute.name == "":
                                resident = data_attribute.resident
                                break

                        size = get_record_size(record)

                    for path in record.full_paths():
                        path = f"{drive_letter}{path}"
                        yield from aggr(
                            self.mft_records(
                                drive_letter=drive_letter,
                                record=record,
                                segment=record.segment,
                                path=path,
                                owner=owner,
                                size=size,
                                resident=resident,
                                inuse=inuse,
                                volume_uuid=volume_uuid,
                                record_formatter=record_formatter,
                            )
                        )
                except Exception as e:
                    self.target.log.warning("An error occured parsing MFT segment %d: %s", record.segment, str(e))
                    self.target.log.debug("", exc_info=e)

        except Exception:
            log.exception("An error occured constructing FilesystemRecords")

    def mft_records(
        self,
        drive_letter: str,
        record: MftRecord,
        segment: int,
        path: str,
        owner: str,
        size: int,
        resident: bool,
        inuse: bool,
        volume_uuid: str,
        record_formatter: Callable,
    ) -> Iterator[Record]:
        for attr in record.attributes.STANDARD_INFORMATION:
            yield from record_formatter(
                attr=attr,
                record_type=InformationType.STANDARD_INFORMATION,
                segment=segment,
                path=windows_path(path),
                owner=owner,
                filesize=size,
                resident=resident,
                inuse=inuse,
                volume_uuid=volume_uuid,
                _target=self.target,
            )

        for idx, attr in enumerate(record.attributes.FILE_NAME):
            filepath = f"{drive_letter}{attr.full_path()}"

            yield from record_formatter(
                attr=attr,
                record_type=InformationType.FILE_INFORMATION,
                filename_index=idx,
                segment=segment,
                path=windows_path(filepath),
                owner=owner,
                filesize=size,
                resident=resident,
                ads=False,
                inuse=inuse,
                volume_uuid=volume_uuid,
                _target=self.target,
            )

        ads_attributes = (data_attr for data_attr in record.attributes.DATA if data_attr.name != "")
        ads_info = record.attributes.FILE_NAME[0]

        for data_attr in ads_attributes:
            resident = data_attr.resident
            size = get_record_size(record, data_attr.name)
            ads_path = f"{path}:{data_attr.name}"

            yield from record_formatter(
                attr=ads_info,
                record_type=InformationType.FILE_INFORMATION,
                filename_index=None,
                segment=segment,
                path=windows_path(ads_path),
                owner=owner,
                filesize=size,
                resident=resident,
                inuse=inuse,
                ads=True,
                volume_uuid=volume_uuid,
                _target=self.target,
            )


def compacted_formatter(attr: Attribute, record_type: InformationType, **kwargs) -> Iterator[Record]:
    record_desc = COMPACT_RECORD_TYPES.get(record_type)
    yield record_desc(
        creation_time=attr.creation_time,
        last_modification_time=attr.last_modification_time,
        last_change_time=attr.last_change_time,
        last_access_time=attr.last_access_time,
        **kwargs,
    )


def formatter(attr: Attribute, record_type: InformationType, **kwargs) -> Iterator[Record]:
    record_desc = RECORD_TYPES.get(record_type)
    for type, timestamp in [
        ("B", attr.creation_time),
        ("C", attr.last_change_time),
        ("M", attr.last_modification_time),
        ("A", attr.last_access_time),
    ]:
        yield record_desc(ts=timestamp, ts_type=type, **kwargs)


def macb_aggr(records: list[Record]) -> Iterator[Record]:
    def macb_set(bitfield, index, letter):
        return bitfield[:index] + letter + bitfield[index + 1 :]

    macbs = []
    for record in records:
        found = False

        offset = 0
        if not getattr(record, "ads", False):
            offset = int(record._desc.name == "filesystem/ntfs/mft/filename") * 5

        field = "MACB".find(record.ts_type) + offset
        for macb in macbs:
            if macb.ts == record.ts and macb.path == record.path:
                macb.macb = macb_set(macb.macb, field, record.ts_type)
                found = True
                break

        if found:
            continue

        macb = FilesystemMACBRecord.init_from_record(record)
        macb.macb = "..../...."
        macb.macb = macb_set(macb.macb, field, record.ts_type)

        macbs.append(macb)

    yield from macbs



from dissect.target.helpers.regutil import VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.trusteddocs import TrustedDocumentsPlugin


def test_trusteddocs_plugin(target_win_users, hive_hku):
    trusteddocs_key_name = "Software\\Microsoft\\Office\\16.0\\Word\\Security\\Trusted Documents"
    trusteddocs_key = VirtualKey(hive_hku, trusteddocs_key_name)

    subkey_name = "TrustRecords"
    subkey = VirtualKey(hive_hku, subkey_name)
    subkey.add_value("c:\\Users\\John\\test", VirtualValue(hive_hku, "c:\\Users\\John\\test", b"test"))

    trusteddocs_key.add_subkey(subkey_name, subkey)
    hive_hku.map_key(trusteddocs_key_name, trusteddocs_key)

    target_win_users.add_plugin(TrustedDocumentsPlugin)

    results = list(target_win_users.trusteddocs())

    assert len(results) == 1

    result = results[0]

    assert result.value == b"test"
    assert result.application == "Word"
    assert result.document_path == "c:\\Users\\John\\test"



from typing import Optional

from dissect.target.filesystem import Filesystem
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.target import Target


class RedHatPlugin(LinuxPlugin):
    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Optional[Filesystem]:
        # also applicable to centos (which is a red hat derivative)
        for fs in target.filesystems:
            if fs.exists("/etc/sysconfig/network-scripts"):
                return fs

        return None



from typing import Iterator
from unittest.mock import patch

import pytest
from asn1crypto.algos import EncryptionAlgorithmId
from flow.record.fieldtypes import datetime as dt

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers import keychain
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugins.apps.browser.firefox import (
    CKA_ID,
    FirefoxPlugin,
    decrypt,
    query_master_key,
    retrieve_master_key,
)
from tests._utils import absolute_path


@pytest.fixture
def target_firefox_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Iterator[Target]:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Mozilla\\Firefox\\Profiles\\g1rbw8y7.default-release\\",
        absolute_path("_data/plugins/apps/browser/firefox/"),
    )

    target_win_users.add_plugin(FirefoxPlugin)

    yield target_win_users


@pytest.fixture
def target_firefox_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Iterator[Target]:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/", absolute_path("_data/plugins/apps/browser/firefox/")
    )
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/default/"),
    )

    target_unix_users.add_plugin(FirefoxPlugin)

    yield target_unix_users


@pytest.fixture
def target_firefox_oculus(target_android: Target, fs_android: VirtualFilesystem) -> Iterator[Target]:
    fs_android.map_dir(
        "/data/data/org.mozilla.vrbrowser",
        absolute_path("_data/plugins/apps/browser/firefox/android/org.mozilla.vrbrowser"),
    )

    target_android.add_plugin(FirefoxPlugin)

    yield target_android


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_history(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.firefox.history())

    assert len(records) == 24
    assert set(["firefox"]) == set(record.browser for record in records)

    assert records[0].url == "https://www.mozilla.org/privacy/firefox/"
    assert records[0].id == "1"
    assert records[0].description == "47356411089529"
    assert records[0].visit_count == 1
    assert records[0].ts == dt("2021-12-01T10:42:05.742000+00:00")


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.firefox.downloads())

    assert len(records) == 3
    assert set(["firefox"]) == set(record.browser for record in records)

    assert records[0].id == 1
    assert records[0].ts_start == dt("2021-12-01T10:57:01.175000+00:00")
    assert records[0].ts_end == dt("2021-12-01T10:57:01.321000+00:00")
    assert (
        records[0].url
        == "https://dl.google.com/tag/s/appguid%3D%7B8A69D345-D564-463C-AFF1-A69D9E530F96%7D%26iid%3D%7B2098EF96-29DB"
        "-B268-0B90-01AD59CD5C17%7D%26lang%3Dnl%26browser%3D3%26usagestats%3D1%26appname%3DGoogle%2520Chrome%26needs"
        "admin%3Dprefers%26ap%3Dx64-stable-statsdef_1%26installdataindex%3Dempty/update2/installers/ChromeSetup.exe"
    )


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_cookies(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.cookies())

    assert len(records) == 4
    assert set(["firefox"]) == set(record.browser for record in records)

    assert records[0].ts_created == dt("2023-07-13 09:53:47.460676+00:00")
    assert sorted([*map(lambda c: c.name, records)]) == [
        "_lr_env_src_ats",
        "_lr_retry_request",
        "_uc_referrer",
        "_uc_referrer",
    ]


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.extensions())

    assert set(["firefox"]) == set(record.browser for record in records)
    assert len(records) == 2
    assert records[0].id == "uBlock0@raymondhill.net"
    assert records[0].ts_install == dt("2024-04-23 07:07:21.516000+00:00")
    assert records[0].ts_update == dt("2024-04-23 07:07:21.516000+00:00")
    assert (
        records[0].ext_path == "C:\\Users\\Win11\\AppData\\Roaming\\Mozilla\\Firefox\\Profiles"
        "\\9nxit8q0.default-release\\extensions\\uBlock0@raymondhill.net.xpi"
    )
    assert records[0].permissions == [
        "alarms",
        "dns",
        "menus",
        "privacy",
        "storage",
        "tabs",
        "unlimitedStorage",
        "webNavigation",
        "webRequest",
        "webRequestBlocking",
    ]
    assert records[1].permissions == []


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_passwords(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.passwords())
    assert len(records) == 2

    for record in records:
        assert record.browser == "firefox"
        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"


def test_unix_firefox_passwords_with_primary_password(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/"),
    )
    target_unix_users.add_plugin(FirefoxPlugin)

    keychain.register_key(
        keychain.KeyType.PASSPHRASE,
        "PrimaryPassword",
        identifier=None,
        provider="browser",
    )

    records = list(target_unix_users.firefox.passwords())

    assert len(records) == 1

    for record in records:
        assert record.browser == "firefox"
        assert record.username == "root"
        assert record.user_home == "/root"

        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"


def test_firefox_oculus_history(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.history())

    assert len(records) == 191
    assert set(["firefox"]) == set(record.browser for record in records)

    assert records[0].url == "https://webxr.today/"
    assert records[0].id == "1"
    assert records[0].description == "47356570952011"
    assert records[0].ts == dt("2021-11-04 13:29:30.780000+00:00")


def test_firefox_oculus_cookies(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.cookies())

    assert len(records) == 33
    assert set(["firefox"]) == set(record.browser for record in records)

    assert records[0].ts_created == dt("2023-07-06 07:50:55.352988+00:00")
    assert sorted([*map(lambda c: c.name, records)]) == [
        "AEC",
        "AWSALBTGCORS",
        "CONSENT",
        "Features",
        "Locale",
        "NID",
        "OptanonConsent",
        "SNID",
        "SOCS",
        "Session-Id",
        "Tag",
        "Tag",
        "Tag",
        "Tag",
        "VISITOR_INFO1_LIVE",
        "_ga",
        "_ga_18EC69JQ0P",
        "_ga_2VC139B3XV",
        "_gat_gtag_UA_36116321_2",
        "_gid",
        "_hjFirstSeen",
        "count",
        "is-proton-user",
        "lst",
        "lst",
        "mp_abe3945ad0ddaadc3d987393d8d7c2ce_mixpanel",
        "session_id",
        "spid.6b32",
        "spses.6b32",
        "visid_incap_1329355",
        "visid_incap_2188750",
        "visid_incap_2295456",
        "visid_incap_2809573",
    ]


def test_firefox_oculus_extensions(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.extensions())

    assert set(["firefox"]) == set(record.browser for record in records)
    assert len(records) == 5
    assert records[0].id == "default-theme@mozilla.org"
    assert records[0].ts_install == dt("2021-11-04 13:29:29.988000+00:00")
    assert records[0].ts_update == dt("1970-01-01 00:00:00+00:00")
    assert records[0].ext_path is None
    assert records[0].permissions == []


@pytest.fixture
def path_key4(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


PRIMARY_PASSWORD = "PrimaryPassword"


@pytest.fixture
def path_key4_primary_password(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


@pytest.fixture
def logins() -> dict[str, str]:
    return {
        "username": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECIFVMX6MyYxpBBC2j8K+bCEaE9/FmqE1wo2A",
        "password": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECB2ZPCZBORUJBBDZ8dBZUVgECoiFD5vPvTbP",
    }


@pytest.fixture
def logins_with_primary_password() -> dict[str, str]:
    return {
        "username": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECPEZvPh6dhBkBBB3/O2puJy1NiBUo5gS8hZh",
        "password": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECM00msnMxFyVBBByTarrnS+FSR5OHQhZfs8t",
    }


@pytest.fixture
def decrypted() -> dict[str, str]:
    return {
        "username": "username",
        "password": "password",
    }


def test_decrypt_is_succesful(path_key4: TargetPath, logins: dict[str, str], decrypted: dict[str, str]) -> None:
    dec_username, dec_password = decrypt(logins.get("username"), logins.get("password"), path_key4)
    assert dec_username == decrypted.get("username")
    assert dec_password == decrypted.get("password")


@patch("dissect.target.plugins.apps.browser.firefox.retrieve_master_key", side_effect=ValueError(""))
def test_decrypt_bad_master_key(path_key4: TargetPath, logins: dict[str, str]) -> None:
    with pytest.raises(ValueError):
        decrypt(logins.get("username"), logins.get("password"), path_key4)


def test_decrypt_with_primary_password_is_succesful(
    path_key4_primary_password: TargetPath, logins_with_primary_password: dict[str, str], decrypted: dict[str, str]
) -> None:
    dec_username, dec_password = decrypt(
        logins_with_primary_password.get("username"),
        logins_with_primary_password.get("password"),
        path_key4_primary_password,
        PRIMARY_PASSWORD,
    )
    assert dec_username == decrypted.get("username")
    assert dec_password == decrypted.get("password")


def test_decrypt_with_bad_primary_password_is_unsuccesful(
    path_key4_primary_password: TargetPath,
    logins_with_primary_password: dict[str, str],
) -> None:
    with pytest.raises(ValueError) as e:
        decrypt(
            logins_with_primary_password.get("username"),
            logins_with_primary_password.get("password"),
            path_key4_primary_password,
            "BAD_PRIMARY_PASSWORD",
        )
        assert e.msg == "Failed to decrypt password using keyfile /key4.db and password 'BAD_PRIMARY_PASSWORD'"


def test_retrieve_master_key_is_succesful(path_key4: TargetPath) -> None:
    key, algorithm = retrieve_master_key(b"", path_key4)

    assert EncryptionAlgorithmId.map(algorithm) == "pbes2"
    assert key.hex() == "452c2f920285f794614af1c2d99ed331940d73298a526140"


@patch("dissect.target.plugins.apps.browser.firefox.query_master_key", return_value=(b"aaaa", "BAD_CKA_VALUE"))
def test_retrieve_master_key_bad_cka_value(mock_target: Target, path_key4: TargetPath) -> None:
    with pytest.raises(ValueError) as e:
        retrieve_master_key(b"", path_key4)
        assert "Password master key CKA_ID 'BAD_CKA_VALUE' is not equal to expected value" in e.msg


def test_query_master_key(path_key4: TargetPath) -> None:
    master_key, master_key_cka = query_master_key(path_key4)

    assert isinstance(master_key, bytes)
    assert len(master_key) == 148

    assert isinstance(master_key_cka, bytes)
    assert len(master_key_cka) == 16
    assert master_key_cka == CKA_ID



from __future__ import annotations

import json
import textwrap
from typing import Iterator, Type

from dissect.target import plugin
from dissect.target.helpers.docs import INDENT_STEP, get_plugin_overview
from dissect.target.plugin import Plugin, arg, export


def categorize_plugins(plugins_selection: list[dict] = None) -> dict:
    """Categorize plugins based on the module it's from."""

    output_dict = dict()

    plugins_selection = plugins_selection or get_exported_plugins()

    for plugin_dict in plugins_selection:
        tmp_dict = dictify_module_recursive(
            list_of_items=plugin_dict["module"].split("."),
            last_value=plugin.load(plugin_dict),
        )
        update_dict_recursive(output_dict, tmp_dict)

    return output_dict


def get_exported_plugins() -> list:
    """Returns list of exported plugins."""
    return [p for p in plugin.plugins() if len(p["exports"])]


def dictify_module_recursive(list_of_items: list, last_value: Plugin) -> dict:
    """Create a dict from a list of strings.

    The last element inside the list, will point to `last_value`
    """
    if len(list_of_items) == 1:
        return {list_of_items[0]: last_value}
    else:
        return {list_of_items[0]: dictify_module_recursive(list_of_items[1:], last_value)}


def update_dict_recursive(source_dict: dict, updated_dict: dict) -> dict:
    """Update source dictionary with data in updated_dict."""

    for key, value in updated_dict.items():
        if isinstance(value, dict):
            source_dict[key] = update_dict_recursive(source_dict.get(key, {}), value)
        else:
            source_dict[key] = value
    return dict(sorted(source_dict.items()))


def output_plugin_description_recursive(
    structure_dict: dict | Plugin,
    print_docs: bool,
    indentation_step: int = 0,
) -> list[str]:
    """Create plugin overview with identations."""

    if isinstance(structure_dict, type) and issubclass(structure_dict, Plugin):
        return [get_plugin_description(structure_dict, print_docs, indentation_step)]

    return get_description_dict(structure_dict, print_docs, indentation_step)


def get_plugin_description(
    plugin_class: Type[Plugin],
    print_docs: bool,
    indentation_step: int,
) -> str:
    """Returns plugin_overview with specific indentation."""

    plugin_overview = get_plugin_overview(
        plugin_class=plugin_class,
        with_func_docstrings=print_docs,
        with_plugin_desc=print_docs,
    )
    return textwrap.indent(plugin_overview, prefix=" " * indentation_step)


def get_description_dict(
    structure_dict: dict,
    print_docs: bool,
    indentation_step: int,
) -> list[str]:
    """Returns a list of indented descriptions."""

    output_descriptions = []
    for key in structure_dict.keys():
        output_descriptions += [
            textwrap.indent(key + ":", prefix=" " * indentation_step) if key != "" else "OS plugins"
        ] + output_plugin_description_recursive(
            structure_dict[key],
            print_docs,
            indentation_step=indentation_step + 2,
        )

    return output_descriptions


class PluginListPlugin(Plugin):
    """Plugin list plugin (so meta)."""

    def check_compatible(self) -> None:
        pass

    @export(output="none", cache=False)
    @arg("--docs", dest="print_docs", action="store_true")
    # NOTE: We would prefer to re-use arguments across plugins from argparse in query.py, but that is not possible yet.
    # For now we use --as-json, but in the future this should be changed to inherit --json from target-query.
    # https://github.com/fox-it/dissect.target/pull/841
    # https://github.com/fox-it/dissect.target/issues/889
    @arg("--as-json", dest="as_json", action="store_true")
    def plugins(self, plugins: list[Plugin] = None, print_docs: bool = False, as_json: bool = False) -> None:
        """Print all available plugins."""

        dict_plugins = list({p.path: p.plugin_desc for p in plugins}.values())
        categorized_plugins = dict(sorted(categorize_plugins(dict_plugins).items()))

        plugin_descriptions = output_plugin_description_recursive(categorized_plugins, print_docs)

        plugins_list = textwrap.indent(
            "\n".join(plugin_descriptions) if plugin_descriptions else "None",
            prefix=INDENT_STEP,
        )

        failed_descriptions = []
        failed_items = plugin.failed()
        for failed_item in failed_items:
            module = failed_item["module"]
            exception = failed_item["stacktrace"][-1].rstrip()
            failed_descriptions.append(
                textwrap.dedent(
                    f"""\
                        Module: {module}
                        Reason: {exception}
                    """
                )
            )

        failed_list = textwrap.indent(
            "\n".join(failed_descriptions) if failed_descriptions else "None",
            prefix=INDENT_STEP,
        )

        output_lines = [
            "Available plugins:",
            plugins_list,
            "",
            "Failed to load:",
            failed_list,
        ]

        if as_json:
            out = {"loaded": list(generate_plugins_json(plugins))}

            if failed_plugins := plugin.failed():
                out["failed"] = [
                    {"module": p["module"], "stacktrace": "".join(p["stacktrace"])} for p in failed_plugins
                ]

            print(json.dumps(out), end="")

        else:
            print("\n".join(output_lines))


def generate_plugins_json(plugins: list[Plugin]) -> Iterator[dict]:
    """Generates JSON output of a list of :class:`Plugin`s."""

    for p in plugins:
        func = getattr(p.class_object, p.method_name)
        description = getattr(func, "__doc__", None)
        summary = description.split("\n\n", 1)[0].strip() if description else None

        yield {
            "name": p.name,
            "output": p.output_type,
            "description": summary,
            "path": p.path,
        }



from typing import Iterator, Union

from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import NamespacePlugin, export

WebserverAccessLogRecord = TargetRecordDescriptor(
    "application/log/webserver/access",
    [
        ("datetime", "ts"),
        ("string", "remote_user"),
        ("net.ipaddress", "remote_ip"),
        ("net.ipaddress", "local_ip"),
        ("varint", "pid"),
        ("string", "method"),
        ("uri", "uri"),
        ("string", "protocol"),
        ("varint", "status_code"),
        ("varint", "bytes_sent"),
        ("uri", "referer"),
        ("string", "useragent"),
        ("varint", "response_time_ms"),
        ("path", "source"),
    ],
)

WebserverErrorLogRecord = TargetRecordDescriptor(
    "application/log/webserver/error",
    [
        ("datetime", "ts"),
        ("net.ipaddress", "remote_ip"),
        ("varint", "pid"),
        ("string", "module"),
        ("string", "level"),
        ("string", "error_source"),
        ("string", "error_code"),
        ("string", "message"),
        ("path", "source"),
    ],
)


class WebserverPlugin(NamespacePlugin):
    __namespace__ = "webserver"
    __findable__ = False

    @export(record=[WebserverAccessLogRecord, WebserverErrorLogRecord])
    def logs(self) -> Iterator[Union[WebserverAccessLogRecord, WebserverErrorLogRecord]]:
        """Returns log file records from installed webservers."""
        yield from self.access()
        yield from self.error()



import textwrap
from datetime import datetime, timedelta, timezone
from io import BytesIO

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.apps.webserver.apache import (
    LOG_FORMAT_ACCESS_COMBINED,
    LOG_FORMAT_ACCESS_COMMON,
    LOG_FORMAT_ACCESS_VHOST_COMBINED,
    ApachePlugin,
)
from dissect.target.target import Target
from tests._utils import absolute_path


def test_infer_access_log_format_combined() -> None:
    log_combined = (
        '127.0.0.1 - - [19/Dec/2022:17:25:12 +0100] "GET / HTTP/1.1" 304 247 "-" "Mozilla/5.0 '
        "(Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 "
        'Safari/537.36"'
    )
    assert ApachePlugin.infer_access_log_format(log_combined) == LOG_FORMAT_ACCESS_COMBINED


def test_infer_access_log_format_vhost_combined() -> None:
    log_vhost_combined = (
        'example.com:80 127.0.0.1 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312 '
        '"-" "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) '
        'Chrome/108.0.0.0 Safari/537.36"'
    )
    assert ApachePlugin.infer_access_log_format(log_vhost_combined) == LOG_FORMAT_ACCESS_VHOST_COMBINED


def test_infer_access_log_format_common() -> None:
    log_common = '127.0.0.1 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312'
    assert ApachePlugin.infer_access_log_format(log_common) == LOG_FORMAT_ACCESS_COMMON


def test_infer_access_log_ipv6() -> None:
    log_combined = (
        '2001:0db8:85a3:0000:0000:8a2e:0370:7334 - - [20/Dec/2022:15:18:01 +0100] "GET / HTTP/1.1" 200 '
        '1126 "-" "curl/7.81.0"'
    )
    assert ApachePlugin.infer_access_log_format(log_combined) == LOG_FORMAT_ACCESS_COMBINED


def test_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6


def test_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log.gz")
    fs_unix.map_file("var/log/apache2/access.log.1.gz", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 4


def test_access_bz2(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log.bz2")
    fs_unix.map_file("var/log/apache2/access.log.1.bz2", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 4


def test_xampp_unix(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("opt/lampp/logs/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6


def test_xampp_windows(target_win: Target, fs_win: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_win.map_file("xampp/apache/logs/access.log", data_file)

    target_win.add_plugin(ApachePlugin)
    results = list(target_win.apache.access())

    assert len(results) == 6


def test_all_access_log_formats(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    tz = timezone(timedelta(hours=1))
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6

    combined_log = results[0]
    assert combined_log.ts == datetime(2022, 12, 19, 17, 6, 19, tzinfo=tz)
    assert combined_log.status_code == 200
    assert combined_log.remote_ip == "1.2.3.4"
    assert combined_log.remote_user == "-"
    assert combined_log.method == "GET"
    assert combined_log.uri == "/"
    assert combined_log.protocol == "HTTP/1.1"
    assert combined_log.referer == "Sample referer"
    assert (
        combined_log.useragent
        == "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 "
        "Safari/537.36"
    )

    vhost_combined_log = results[1]
    assert vhost_combined_log.ts == datetime(2022, 12, 19, 17, 25, 40, tzinfo=tz)
    assert vhost_combined_log.status_code == 200
    assert vhost_combined_log.remote_ip == "1.2.3.4"
    assert vhost_combined_log.remote_user == "-"
    assert vhost_combined_log.method == "GET"
    assert vhost_combined_log.uri == "/index.html"
    assert vhost_combined_log.protocol == "HTTP/1.1"
    assert vhost_combined_log.referer == "-"
    assert (
        vhost_combined_log.useragent
        == "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 "
        "Safari/537.36"
    )

    common_log = results[2]
    assert common_log.ts == datetime(2022, 12, 19, 17, 25, 48, tzinfo=tz)
    assert common_log.status_code == 200
    assert common_log.remote_ip == "4.3.2.1"
    assert common_log.remote_user == "-"
    assert common_log.method == "GET"
    assert common_log.uri == "/"
    assert common_log.protocol == "HTTP/1.1"
    assert common_log.referer is None
    assert common_log.useragent is None

    ipv6_log = results[3]
    assert ipv6_log.remote_ip == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    assert ipv6_log.protocol is None
    assert ipv6_log.bytes_sent == 0


def test_logrotate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)
    fs_unix.map_file("var/log/apache2/access.log.1", data_file)
    fs_unix.map_file("var/log/apache2/access.log.2", data_file)
    fs_unix.map_file("var/log/apache2/access.log.3", data_file)

    target_unix.add_plugin(ApachePlugin)
    access_log_paths, error_log_paths = target_unix.apache.get_log_paths()

    assert len(access_log_paths) == 4
    assert len(error_log_paths) == 0


def test_custom_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("etc/apache2/apache2.conf", BytesIO(b'CustomLog "/custom/log/location/access.log" common'))
    fs_unix.map_file_fh("custom/log/location/access.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("custom/log/location/access.log.1", BytesIO(b"Foo1"))
    fs_unix.map_file_fh("custom/log/location/access.log.2", BytesIO(b"Foo2"))
    fs_unix.map_file_fh("custom/log/location/access.log.3", BytesIO(b"Foo3"))

    target_unix.add_plugin(ApachePlugin)
    access_log_paths, error_log_paths = target_unix.apache.get_log_paths()

    assert len(access_log_paths) == 4
    assert len(error_log_paths) == 0


def test_config_commented_logs(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    # CustomLog "/custom/log/location/old.log" common
    CustomLog "/custom/log/location/new.log" common
    # ErrorLog "/custom/log/location//old_error.log"
    ErrorLog "/custom/log/location//new_error.log"

    """
    fs_unix.map_file_fh("etc/httpd/conf/httpd.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("custom/log/location/new.log", BytesIO(b"New"))
    fs_unix.map_file_fh("custom/log/location/old.log", BytesIO(b"Old"))
    fs_unix.map_file_fh("custom/log/location/old_error.log", BytesIO(b"Old"))
    fs_unix.map_file_fh("custom/log/location/new_error.log", BytesIO(b"New"))
    target_unix.add_plugin(ApachePlugin)

    access_log_paths, error_log_paths = target_unix.apache.get_log_paths()

    # Log paths are returned in alphabetical order
    assert str(access_log_paths[0]) == "/custom/log/location/new.log"
    assert str(access_log_paths[1]) == "/custom/log/location/old.log"

    assert str(error_log_paths[0]) == "/custom/log/location/new_error.log"
    assert str(error_log_paths[1]) == "/custom/log/location/old_error.log"


def test_error_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/error.log")
    fs_unix.map_file("var/log/apache2/error.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.error())

    assert len(results) == 3
    assert all(str(record.source) == "/var/log/apache2/error.log" for record in results)


def test_all_error_log_formats(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/error.log")
    fs_unix.map_file("var/log/apache2/error.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.error())

    assert len(results) == 3

    no_client_log = results[0]
    assert no_client_log.ts == datetime(2022, 12, 13, 13, 7, 21, 131489, tzinfo=timezone.utc)
    assert no_client_log.error_code == "AH00163"
    assert no_client_log.module == "mpm_prefork"
    assert no_client_log.level == "notice"
    assert no_client_log.pid == 218
    assert no_client_log.message == (
        "Apache/2.4.54 (Unix) mod_perl/2.0.10 Perl/v5.30.2 configured -- resuming normal operations"
    )

    client_log = results[1]
    assert client_log.ts == datetime(2011, 5, 12, 8, 28, 57, 652118, tzinfo=timezone.utc)
    assert client_log.error_code is None
    assert client_log.remote_ip == "::1"
    assert client_log.module == "core"
    assert client_log.level == "error"
    assert client_log.pid == 8777
    assert client_log.message == "File does not exist: /usr/local/apache2/htdocs/favicon.ico"

    log_with_source = results[2]
    assert log_with_source.ts == datetime(2019, 2, 1, 22, 3, 8, 320285, tzinfo=timezone.utc)
    assert log_with_source.error_code == "AH01626"
    assert log_with_source.remote_ip == "172.17.0.1"
    assert log_with_source.module == "authz_core"
    assert log_with_source.level == "debug"
    assert log_with_source.pid == 7
    assert log_with_source.message == "authorization result of <RequireAny>: granted"
    assert log_with_source.error_source == "mod_authz_core.c(820)"



import re
from enum import Enum, auto
from typing import Optional, Tuple
from uuid import UUID

from dissect.ntfs.exceptions import FileNotFoundError
from dissect.ntfs.mft import MftRecord

from dissect.target import Target
from dissect.target.filesystems.ntfs import NtfsFilesystem

DRIVE_LETTER_RE = re.compile(r"[a-zA-Z]:")


class InformationType(Enum):
    """Valid information types"""

    STANDARD_INFORMATION = auto()
    FILE_INFORMATION = auto()
    ALTERNATE_DATA_STREAM = auto()


def get_drive_letter(target: Target, filesystem: NtfsFilesystem) -> str:
    """Retrieve the drive letter from the loaded mounts

    When the drive letter is not available for that filesystem it returns empty.
    """
    # A filesystem can be known under multiple drives (mount points). If it is
    # a windows system volume, there are the default sysvol and c: drives.
    # If the target has a virtual ntfs filesystem, e.g. as constructed by the
    # tar and dir loaders, there is also the /$fs$/fs<n> drive, under which the
    # "fake" ntfs filesystem is mounted.
    # The precedence for drives is first the drive letter drives (e.g. c:),
    # second the "normally" named drives (e.g. sysvol) and finally the anonymous
    # drives (e.g. /$fs/fs0).
    mount_items = (item for item in target.fs.mounts.items() if hasattr(item[1], "ntfs"))
    drives = [key for key, fs in mount_items if fs.ntfs is filesystem.ntfs]

    single_letter_drives = []
    other_drives = []
    anon_drives = []

    for drive in drives:
        if DRIVE_LETTER_RE.match(drive):
            single_letter_drives.append(drive)
        elif "$fs$" in drive:
            anon_drives.append(drive)
        else:
            other_drives.append(drive)

    drives = sorted(single_letter_drives) + sorted(other_drives) + sorted(anon_drives)

    if drives:
        return f"{drives[0]}\\"
    else:
        return ""


def get_volume_identifier(fs: NtfsFilesystem) -> str:
    """Return the filesystem guid if available"""
    try:
        return f"{UUID(bytes_le=fs.volume.guid)}"
    except (AttributeError, TypeError, ValueError):
        # AttributeError is raised when volume is None
        # TypeError is raised when guid is None
        # ValueError is raised when the guid string is smaller than 16 bytes
        return None


def get_owner_and_group(entry: MftRecord, fs: NtfsFilesystem) -> Tuple[Optional[str], Optional[str]]:
    owner, group = None, None
    try:
        stdinfo = entry.attributes.STANDARD_INFORMATION[0]
        sd = fs.ntfs.secure.lookup(stdinfo.security_id)
        owner, group = sd.owner, sd.group
    except (AttributeError, IndexError, KeyError):
        # When $STANDARD_INFORMATION cannot be found or no $Secure file is loaded
        pass
    return owner, group


def get_record_size(record: MftRecord, name: str = "") -> Optional[int]:
    """Gets the size for a specific record"""
    try:
        return record.size(name)
    except (FileNotFoundError, KeyError):
        # FileNotFoundError is raised when the entry has no attributes
        # KeyError is raised when it tries to access a specific attribute that does not exist
        pass
    return None



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import Mock, mock_open, patch

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.recyclebin import RecyclebinPlugin, c_recyclebin


@pytest.fixture
def recycle_bin():
    recycle_bin = VirtualFilesystem()

    with TemporaryDirectory() as tmp_dir:
        recycle_bin.map_dir("$a_random_recycle_bin_file", tmp_dir)
        yield recycle_bin


def test_recycle_bin_compatibility_failed(target_win):
    with pytest.raises(UnsupportedPluginError):
        RecyclebinPlugin(target_win).check_compatible()


def test_recycle_bin_compat_succeeded(target_win, recycle_bin):
    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    assert RecyclebinPlugin(target_win).check_compatible() is None


def test_read_recycle_bin(target_win):
    mocked_file = Mock()

    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False
    with patch.object(RecyclebinPlugin, "read_bin_file") as mocked_bin_file:
        assert [mocked_bin_file.return_value] == list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_file))


def test_filtered_name(target_win):
    mocked_file = Mock()
    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False

    mocked_file.name = "hello"

    assert [] == list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_file))


def test_read_recycle_bin_directory(target_win):
    mocked_dir = Mock()
    mocked_dir.is_file.return_value = False
    mocked_dir.is_dir.return_value = True
    mocked_dir.name = "SID-SOME_RANDOM_SID"

    mocked_file = Mock()
    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False
    mocked_file.name = "$ihello"

    mocked_dir.iterdir.return_value = [mocked_file] * 3

    with patch.object(RecyclebinPlugin, "read_bin_file", return_value=mocked_file):
        data = list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_dir))

        assert data == [mocked_file] * 3


@pytest.mark.parametrize(
    "version_number, expected_header",
    [
        (b"\x00" * 8, "header_v1"),
        (b"\x01" * 8, "header_v1"),
        (b"\x02" + b"\x00" * 7, "header_v2"),
    ],
)
def test_parse_header(target_win, version_number, expected_header):
    recycle_plugin = RecyclebinPlugin(target_win)

    header = recycle_plugin.select_header(version_number)
    assert header is getattr(c_recyclebin, expected_header)


@pytest.mark.parametrize(
    "path",
    [
        "file/to/some/sid",
        "file/to/$recycle.bin/sid",
    ],
)
def test_read_bin_file_unknown(target_win, path):
    recycle_plugin = RecyclebinPlugin(target_win)

    header_1 = c_recyclebin.header_v1(version=0, file_size=0x20, timestamp=0x20, filename="hello_world" + "\x00" * 249)

    with patch.object(Path, "open", mock_open(read_data=header_1.dumps())):
        normal_path = Path(path)

        output = recycle_plugin.read_bin_file(normal_path)

    assert output.filesize == 0x20
    assert output.path == "hello_world"


def test_recyclebin_plugin_file(target_win, recycle_bin):
    recycle_bin.map_file("$ihello_world", None)

    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    target_win.add_plugin(RecyclebinPlugin)

    with patch.object(RecyclebinPlugin, "read_bin_file") as mocked_bin_file:
        recycle_bin_entries = list(target_win.recyclebin())
        assert recycle_bin_entries == [mocked_bin_file.return_value]


def test_recyclebin_plugin_wrong_prefix(target_win, recycle_bin):
    recycle_bin.map_file("hello_world", None)
    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    target_win.add_plugin(RecyclebinPlugin)

    with patch.object(RecyclebinPlugin, "read_bin_file"):
        recycle_bin_entries = list(target_win.recyclebin())
        assert recycle_bin_entries == []


@pytest.mark.parametrize(
    "path, expected_output",
    [
        ("C:/$Recycle.bin/sid-data/file", "sid-data"),
        ("$Recycle.bin/sid-data/file", "sid-data"),
        ("$Recycle.bin/just_another_file", "unknown"),
        ("C:/$Recycle.bin/just_another_file", "unknown"),
    ],
)
def test_find_sid_from_path(target_win, path, expected_output):
    recycle_plugin = RecyclebinPlugin(target_win)

    assert recycle_plugin.find_sid(Path(path)) == expected_output



"""Filesystem and path related utilities."""

from __future__ import annotations

import fnmatch
import gzip
import hashlib
import io
import logging
import os
import re
import sys
from pathlib import Path
from typing import Any, BinaryIO, Iterator, Optional, Sequence, TextIO, Union

try:
    import lzma

    HAS_XZ = True
except ImportError:
    HAS_XZ = False

try:
    import bz2

    HAS_BZ2 = True
except ImportError:
    HAS_BZ2 = False

try:
    import zstandard

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

import dissect.target.filesystem as filesystem
from dissect.target.exceptions import FileNotFoundError, SymlinkRecursionError
from dissect.target.helpers.polypath import (
    abspath,
    basename,
    commonpath,
    dirname,
    isabs,
    join,
    normalize,
    normpath,
    relpath,
    split,
    splitdrive,
    splitext,
    splitroot,
)

if sys.version_info >= (3, 12):
    from dissect.target.helpers.compat.path_312 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 11):
    from dissect.target.helpers.compat.path_311 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 10):
    from dissect.target.helpers.compat.path_310 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 9):
    from dissect.target.helpers.compat.path_39 import PureDissectPath, TargetPath
else:
    raise RuntimeError("dissect.target requires at least Python 3.9")


log = logging.getLogger(__name__)

re_glob_magic = re.compile(r"[*?[]")
re_glob_index = re.compile(r"(?<=\/)[^\/]*[*?[]")

__all__ = [
    "abspath",
    "basename",
    "commonpath",
    "dirname",
    "fs_attrs",
    "generate_addr",
    "glob_ext",
    "glob_split",
    "has_glob_magic",
    "isabs",
    "join",
    "normalize",
    "normpath",
    "open_decompress",
    "PureDissectPath",
    "relpath",
    "resolve_link",
    "reverse_readlines",
    "split",
    "splitdrive",
    "splitext",
    "splitroot",
    "stat_result",
    "TargetPath",
    "walk_ext",
    "walk",
    "recurse",
]


def generate_addr(path: Union[str, Path], alt_separator: str = "") -> int:
    if not alt_separator and isinstance(path, Path):
        alt_separator = path._flavour.altsep
    path = normalize(str(path), alt_separator=alt_separator)
    return int(hashlib.sha256(path.encode()).hexdigest()[:8], 16)


class stat_result:  # noqa
    """Custom stat_result object, designed to mimick os.stat_result.

    The real stat_result is a CPython internal StructSeq, which kind of behaves like a namedtuple on steroids.
    We try to emulate some of that behaviour here.

    For consistency this class is also called stat_result.
    """

    __slots__ = {
        # Regular fields
        "st_mode": "protection bits",
        "st_ino": "inode",
        "st_dev": "device",
        "st_nlink": "number of hard links",
        "st_uid": "user ID of owner",
        "st_gid": "group ID of owner",
        "st_size": "total size, in bytes",
        # Regular timestamp fields (integer, float, nanosecond)
        "_st_atime": "integer time of last access",
        "_st_mtime": "integer time of last modification",
        "_st_ctime": "integer time of last change",
        "st_atime": "time of last access",
        "st_mtime": "time of last modification",
        "st_ctime": "time of last change",
        "st_atime_ns": "time of last access in nanoseconds",
        "st_mtime_ns": "time of last modification in nanoseconds",
        "st_ctime_ns": "time of last change in nanoseconds",
        # Extra optional fields
        "st_blksize": "blocksize for filesystem I/O",
        "st_blocks": "number of blocks allocated",
        "st_rdev": "device type (if inode device)",
        "st_flags": "user defined flags for file",
        "st_gen": "generation number",
        "st_birthtime": "time of creation",
        "st_file_attributes": "Windows file attribute bits",
        "st_fstype": "Type of filesystem",
        "st_reparse_tag": "Windows reparse tag",
        "st_birthtime_ns": "time of creation in nanoseconds",
        # Internal fields
        "_s": "internal tuple",
    }
    _field_count = len(__slots__) - 1

    def __init__(self, s: Sequence[Any]):
        if not isinstance(s, (list, tuple)) or not len(s) >= 10:
            raise TypeError(f"dissect.target.stat_result() takes an at least 10-sequence ({len(s)}-sequence given)")

        if len(s) > self._field_count:
            raise TypeError(
                f"dissect.target.stat_result() takes an at most {self._field_count}-sequence ({len(s)}-sequence given)"
            )

        s = tuple(s)
        if len(s) < self._field_count:
            s = s + ((None,) * (self._field_count - len(s)))

        self.st_mode = s[0]
        self.st_ino = s[1]
        self.st_dev = s[2]
        self.st_nlink = s[3]
        self.st_uid = s[4]
        self.st_gid = s[5]
        self.st_size = s[6]

        # Parse the values given to us in the "integer" field and calculate
        # the integer, float and nanosecond timestamps from there.
        self._st_atime, self.st_atime, self.st_atime_ns = self._parse_time(s[7])
        self._st_mtime, self.st_mtime, self.st_mtime_ns = self._parse_time(s[8])
        self._st_ctime, self.st_ctime, self.st_ctime_ns = self._parse_time(s[9])

        # Allow overriding the calculated values so a more accurate nanosecond timestamp can be specified
        self.st_atime = s[10] or self.st_atime
        self.st_mtime = s[11] or self.st_mtime
        self.st_ctime = s[12] or self.st_ctime
        self.st_atime_ns = s[13] or self.st_atime_ns
        self.st_mtime_ns = s[14] or self.st_mtime_ns
        self.st_ctime_ns = s[15] or self.st_ctime_ns

        self.st_blksize = s[16]
        self.st_blocks = s[17]
        self.st_rdev = s[18]
        self.st_flags = s[19]
        self.st_gen = s[20]
        self.st_birthtime = s[21]
        self.st_file_attributes = s[22]
        self.st_fstype = s[23]
        self.st_reparse_tag = s[24]
        self.st_birthtime_ns = s[25]

        # stat_result behaves like a tuple, but only with the first 10 fields
        # Note that this means it specifically uses the integer variants of the timestamps
        self._s = (
            self.st_mode,
            self.st_ino,
            self.st_dev,
            self.st_nlink,
            self.st_uid,
            self.st_gid,
            self.st_size,
            self._st_atime,
            self._st_mtime,
            self._st_ctime,
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, stat_result):
            other = other._s

        return self._s == other

    def __ne__(self, other) -> bool:
        return not self == other

    def __getitem__(self, item) -> int:
        return self._s[item]

    def __iter__(self) -> Iterator[int]:
        return iter(self._s)

    def __repr__(self) -> str:
        values = ", ".join(
            f"{k}={getattr(self, k)}" for k in self.__slots__ if k.startswith("st_") and getattr(self, k) is not None
        )
        return f"dissect.target.stat_result({values})"

    def _parse_time(self, ts: Union[int, float]) -> tuple[int, float, int]:
        ts_int = int(ts)
        ts_ns = int(ts * 1e9)

        return ts_int, ts_ns * 1e-9, ts_ns

    @classmethod
    def copy(cls, other) -> stat_result:
        # First copy the basic 10 fields
        st = cls(list(other))
        # Then iterate and copy any other
        for attr in list(cls.__slots__.keys())[10 : cls._field_count]:
            try:
                setattr(st, attr, getattr(other, attr))
            except AttributeError:
                pass
        return st


def walk(path_entry, topdown=True, onerror=None, followlinks=False):
    for path_list, dirs, files in walk_ext(path_entry, topdown, onerror, followlinks):
        dir_names = [d.name for d in dirs]
        file_names = [f.name for f in files]

        walk_path = join(path_entry.path, *[p.name for p in path_list[1:]])
        yield walk_path, dir_names, file_names

        if len(dir_names) != len(dirs):
            dirs[:] = [d for d in dirs if d.name in dir_names]


def walk_ext(path_entry, topdown=True, onerror=None, followlinks=False):
    dirs = []
    files = []

    try:
        for entry in path_entry.scandir():
            if entry.is_dir():
                dirs.append(entry)
            else:
                files.append(entry)
    except Exception as e:
        if onerror is not None and callable(onerror):
            e.entry = path_entry
            onerror(e)
        return

    if topdown:
        yield [path_entry], dirs, files

    for direntry in dirs:
        if followlinks or not direntry.is_symlink():
            for xpath, xdirs, xfiles in walk_ext(direntry, topdown, onerror, followlinks):
                yield [path_entry] + xpath, xdirs, xfiles

    if not topdown:
        yield [path_entry], dirs, files


def recurse(path_entry: filesystem.FilesystemEntry) -> Iterator[filesystem.FilesystemEntry]:
    """Recursively walk the given :class:`FilesystemEntry`, yields :class:`FilesystemEntry` instances."""
    yield path_entry

    if not path_entry.is_dir():
        return

    for child_entry in path_entry.scandir():
        if child_entry.is_dir() and not child_entry.is_symlink():
            yield from recurse(child_entry)
        else:
            yield child_entry


def glob_split(pattern: str, alt_separator: str = "") -> tuple[str, str]:
    """Split a pattern on path part boundaries on the first path part with a glob pattern.

    Args:
        pattern: A glob pattern to match names of filesystem entries against.
        alt_separator: An optional alternative path separator in use by the filesystem being matched.

    Returns:
        A tuple of a string with path parts up to the first path part that has a glob pattern and a string of
        the remaining path parts.
    """
    # re_glob_index expects a normalized pattern
    pattern = normalize(pattern, alt_separator=alt_separator)

    first_glob = re_glob_index.search(pattern)

    if not first_glob:
        return pattern, ""

    pos = first_glob.start()
    return pattern[:pos], pattern[pos:]


def glob_ext(direntry: filesystem.FilesystemEntry, pattern: str) -> Iterator[filesystem.FilesystemEntry]:
    """Recursively search and return filesystem entries matching a given glob pattern.

    Args:
        direntry: The filesystem entry relative to which to search.
        pattern: A glob pattern to match names of filesystem entries against.

    Yields:
        Matching filesystem entries (files and/or directories).
    """

    # Split the pattern on the last path part. base_name will contain the last path part (which is
    # '' if pattern ends with a /) and dir_name will contain the other parts.
    dir_name, base_name = split(pattern, alt_separator=direntry.fs.alt_separator)

    # The simple case where there are no globs.
    if not has_glob_magic(pattern):
        try:
            entry = direntry.get(pattern)
        except FileNotFoundError:
            pass
        else:
            # Patterns ending with a slash, so without a base_name, should match only directories.
            if base_name:
                yield entry
            elif entry.is_dir():
                yield entry
        return

    # The pattern has only one path part, so we can match directly against the files in direntry.
    if not dir_name:
        for entry in glob_ext1(direntry, base_name):
            yield entry
        return

    # If the pattern has more than one path part and these parts (dir_name) contain globs, we
    # recursively go over all the path parts and match them (glob_ext).
    # If these path parts have no globs, we get the path directly by name (glob_ext0).
    if has_glob_magic(dir_name):
        glob_in_dir = glob_ext
    else:
        glob_in_dir = glob_ext0

    # If the pattern's last path part (base_name) has globs, we fnmatch it against the entries in
    # direntry (glob_ext1), otherwise we get the part directly by name (glob_ext0).
    if has_glob_magic(base_name):
        glob_in_base = glob_ext1
    else:
        glob_in_base = glob_ext0

    for direntry in glob_in_dir(direntry, dir_name):
        for entry in glob_in_base(direntry, base_name):
            yield entry


# These 2 helper functions non-recursively glob inside a literal directory.
def glob_ext1(direntry: filesystem.FilesystemEntry, pattern: str) -> Iterator[filesystem.FilesystemEntry]:
    """Match and return filesystem entries in a given filesystem entry based on pattern.

    Args:
        direntry: The filesystem entry relative to which to match the entries.
        pattern: A glob pattern to match names of filesystem entries against.

    Yields:
        Matching filesystem entries (files and/or directories).
    """
    if not direntry.is_dir():
        return

    entries = direntry.scandir()

    if pattern[0] != ".":
        # Do not return dot-files, unless they are explicitly searched for.
        entries = filter(lambda x: x.name[0] != ".", entries)

    for entry in entries:
        case_sensitive = entry.fs.case_sensitive
        name = entry.name if case_sensitive else entry.name.lower()
        pattern = pattern if case_sensitive else pattern.lower()
        if fnmatch.fnmatch(name, pattern):
            yield entry


def glob_ext0(direntry: filesystem.FilesystemEntry, path: str) -> Iterator[filesystem.FilesystemEntry]:
    """Return the filesystem entry equal to the given path relative to direntry.

    Args:
        direntry: The filesystem entry relative to which to return the path.
        path: The path to the filesystem entry to return (if present). If path
              is an empty string (``''``) the ``direntry`` itself is returned.

    Yields:
        The matching filesystem entry (file or directory).
    """
    if path == "":
        # os.path.split() returns an empty path for paths ending with a directory separator.
        # E.g. 'q*x/' should match only directories.
        if direntry.is_dir():
            yield direntry
    elif direntry.is_dir():
        try:
            yield direntry.get(path)
        except FileNotFoundError:
            pass


def has_glob_magic(s) -> bool:
    return re_glob_magic.search(s) is not None


def resolve_link(
    fs: filesystem.Filesystem,
    link: str,
    path: str,
    *,
    alt_separator: str = "",
    previous_links: set[str] | None = None,
) -> filesystem.FilesystemEntry:
    """Resolves a symlink to its actual path.

    It stops resolving once it detects an infinite recursion loop.
    """

    link = normalize(link, alt_separator=alt_separator)
    path = normalize(path, alt_separator=alt_separator)

    # Create hash for entry based on path and link
    link_id = f"{path}{link}"
    hash_entry = hash(link_id)

    if not previous_links:
        previous_links = set()

    # Check whether the current entry was already resolved once.
    if hash_entry in previous_links:
        raise SymlinkRecursionError(f"Symlink loop detected for {link_id}")

    previous_links.add(hash_entry)

    if not isabs(link):
        cur_dirname = dirname(normpath(path))
        link = normpath(join(cur_dirname, link))

    # retrieve file from root
    entry = fs.get(link)

    if entry.is_symlink():
        entry = resolve_link(
            fs,
            entry.readlink(),
            link,
            alt_separator=entry.fs.alt_separator,
            previous_links=previous_links,
        )

    return entry


def open_decompress(
    path: Optional[TargetPath] = None,
    mode: str = "rb",
    *,
    fileobj: Optional[BinaryIO] = None,
    encoding: Optional[str] = "UTF-8",
    errors: Optional[str] = "backslashreplace",
    newline: Optional[str] = None,
) -> Union[BinaryIO, TextIO]:
    """Open and decompress a file. Handles gz, bz2 and zstd files. Uncompressed files are opened as-is.

    When passing in an already opened ``fileobj``, the mode, encoding, errors and newline arguments are ignored.

    Args:
        path: The path to the file to open and decompress. It is assumed this path exists.
        mode: The mode in which to open the file.
        fileobj: The file-like object to open and decompress. This is mutually exclusive with path.
        encoding: The decoding for text streams. By default UTF-8 encoding is used.
        errors: The error handling for text streams. By default we're more lenient and use ``backslashreplace``.
        newline: How newlines are handled for text streams.

    Returns:
        An binary or text IO stream, depending on the mode with which the file was opened.

    Example:
        bytes_buf = open_decompress(Path("/dir/file.gz")).read()

        for line in open_decompress(Path("/dir/file.gz"), "rt"):
            print(line)
    """
    if path and fileobj:
        raise ValueError("path and fileobj are mutually exclusive")

    if not path and not fileobj:
        raise ValueError("path or fileobj is required")

    if path:
        file = path.open("rb")
    else:
        file = fileobj

    magic = file.read(5)
    file.seek(0)

    if "b" in mode:
        # Reset the default encoding and errors mode in case of a binary stream
        encoding = None
        errors = None

    if magic[:2] == b"\x1f\x8b":
        return gzip.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if HAS_XZ and magic[:5] == b"\xfd7zXZ":
        return lzma.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if HAS_BZ2 and magic[:3] == b"BZh" and 0x31 <= magic[3] <= 0x39:
        # In a valid bz2 header the 4th byte is in the range b'1' ... b'9'.
        return bz2.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if HAS_ZSTD and magic[:4] in [b"\xfd\x2f\xb5\x28", b"\x28\xb5\x2f\xfd"]:
        # stream_reader is not seekable, so we have to resort to the less
        # efficient decompressor which returns bytes.
        return io.BytesIO(zstandard.decompress(file.read()))

    if path:
        file.close()
        return path.open(mode, encoding=encoding, errors=errors, newline=newline)

    return file


def reverse_readlines(fh: TextIO, chunk_size: int = 1024 * 1024 * 8) -> Iterator[str]:
    """Like iterating over a ``TextIO`` file-like object, but starting from the end of the file.

    Args:
        fh: The file-like object (opened in text mode) to iterate lines from.
        chunk_size: The chunk size to use for iterating over lines.

    Returns:
        An iterator of lines from the file-like object, in reverse.
    """
    offset = fh.seek(0, io.SEEK_END) & ((1 << 64) - 1)
    lines = []

    prev_offset = offset
    while offset > 0:
        if offset < chunk_size:
            chunk_size = offset
        offset -= chunk_size
        fh.seek(offset)

        lines = []
        # tell() on TextIO returns a cookie which includes encode/decoder state
        # Lower 64 bit are the file position
        # https://peps.python.org/pep-3116/#text-i-o
        # See TextIOWrapper._pack_cookie in _pyio.py for more detail
        while (fh.tell() & ((1 << 64) - 1)) < prev_offset:
            try:
                lines.append(fh.readline())
            except UnicodeDecodeError:
                offset += 1
                fh.seek(offset)

        yield from reversed(lines[1:])

        if prev_offset == offset:
            # Previous lines are unreadable due to decoding errors
            raise UnicodeDecodeError(fh.encoding, b"", 0, offset + 1, "failed to decode line")

        prev_offset = offset

    if lines:
        yield lines[0]


def fs_attrs(
    path: Union[os.PathLike, str, bytes],
    follow_symlinks: bool = True,
) -> dict[Union[os.PathLike, str, bytes], bytes]:
    """Return the extended attributes for a given path on the local filesystem.

    This is currently only implemented for Linux using os.listxattr and related functions.

    Args:
        path: The path to get the extended attributes for.
        follow_symlinks: Wether to follow the symlink if the given path is a symlink.

    Returns:
        A dict containing the attribute names as keys and their values.
    """
    attrs = {}
    if hasattr(os, "listxattr"):
        # os.listxattr etc. are only available on Linux
        attr_names = os.listxattr(path, follow_symlinks=follow_symlinks)
        for attr_name in attr_names:
            attrs[attr_name] = os.getxattr(path, attr_name, follow_symlinks=follow_symlinks)

    return attrs
